/*\

    SEMMS_MOD.C

    Program SEMMS (Spatial EM for Markovian Segmentation) : routines calcul parametres classes

    Van Mo DANG       Janvier 96


    Vers-mod  Date         Who Description

    1.03-a    01-NOV-1996  MD  Add function SetIdMatrix + calls by ParaPk.. 
                               to output variance matrix in file don.mf
    1.04-a    08-JAN-1998  MD  Fixed error in DensPkVkI log(vk) needs * Nd
    1.05-a    12-JAN-1998  MD  MissMode in ParaP*V*I
    1.05-b    12-JAN-1998  MD  Process missing data in DensPkVkI
    1.05-c    12-JAN-1998  MD  Add ParaPkV*I_Missing() (empty)
    1.05-d    15-JAN-1998  MD  Add CommonGaussDiagMissing() <= ParaPkV*I_Missing()
    1.05-e    26-JAN-1998  MD  Call to GenAlloc
    1.06-a    29-JUN-1998  MD  Reflect change to nem_mod.h
    1.06-b    29-JUN-1998  MD  Reflect change to nem_typ.h
    1.06-c    29-JUN-1998  MD  Compute Para.NbObs_KD
    1.06-d    02-JUL-1998  MD  Bug: CommonGaus/replace, sum(x-m)^2 neq sum(x^2)-m^2
    1.06-e    02-JUL-1998  MD  Bug: in VkI_Missing/replace, div inertia by nk*D
    1.06-f    02-JUL-1998  MD  Add EstimParaLaplace and called subfunctions
    1.06-g    01-DEC-1998  MD  FkP double* instead of *float in dens...
              01-JUIN-2000 NP  Add model NORMALGEN: full covariance matrix for multivariate
                               gaussian density
\*/

#include "genmemo.h"    /* GenAlloc */
#include "semms_typ.h"    /* DataT, ... */
#include "semms_mod.h"    /* ParaP_V_I, ... */
#include <stdio.h>      /* printf, ... */
#include <stdlib.h>     /* malloc, ... */
#include <string.h>     /* memcpy, ... */
#include <math.h>       /* exp, ... */
#include <values.h>     /* MAXDOUBLE */


#define TWO_PI    2 * 3.14159 

/*V1.05-d*/
#define _IJ       ( ( i * D ) + j )    /* access a (.,D) matrix by (i,j) */
#define _HJ       ( ( h * D ) + j )    /* access a (.,D) matrix by (h,j) */
#define _IH       ( ( i * K ) + h )    /* access a (.,K) matrix by (i,h) */
#define sqr(x)    ((x)*(x))            /* macro for x^2 */



/* ==================== LOCAL FUNCTION PROTOTYPING =================== */


/* Indirect call by GetDensityFunc() */

static int DensNormal      /* ret : 0 if OK, -1 if zero density */
        (
            int                Nd,         /* I : point dimension */
            int                Ik,         /* I : class number : 0..Nk-1 */
            const ModelParaT*  ParaP ,     /* I : model parameters */
            const double*       XV,        /* I : point (dim d) */
            const ModelSpecT*  ModelSpecP,
            double*            FkP,        /* O : density for class Ik */
            double*            LogFkP,      /* O : log of density */
            const double*            InvChol,
            const double            DiagChol 
        ) ;


static int DensLaplaceDiag      /* ret : 0 if OK, -1 if zero density */
        (
            int                Nd,         /* I : point dimension */
            int                Ik,         /* I : class number : 0..Nk-1 */
            const ModelParaT*  ParaP ,     /* I : model parameters */
            const double*       XV,         /* I : point (dim d) */
            const ModelSpecT*  ModelSpecP,
            double*            FkP,        /* O : density for class Ik */
            double*             LogFkP,      /* O : log of density */
            const double*            InvChol,
            const double            DiagChol 
        ) ;



/* Called by EstimPara() */

static StatusET            /* ret : OK, W_EMPTYCLASS or E_MEMORY */ /*V1.06-b*/
EstimParaNormal 
(
  const double   *C_NK,      /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           Nk,         /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/
  const ModelSpecT* SpecP,  /* I : model specification */ /*V1.06-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 */ /*V1.05-b*/
  ModelParaT    *ParaP      /* O : estimated parameters */
) ;




static StatusET            /* ret : OK, W_EMPTYCLASS or E_MEMORY */ /*V1.06-b*/
EstimParaLaplace
(
  const double   *C_NK,      /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           Nk,         /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/
  const ModelSpecT* SpecP,  /* I : model specification */ /*V1.06-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 */ /*V1.05-b*/
  ModelParaT    *ParaP      /* O : estimated parameters */
) ;




/* Called by EstimParaNormal() */

/*V1.05-d*/
static StatusET          /* Return status : OK, EMPTY or MEMORY */
CommonGauss
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const ModelSpecT* SpecP,  /* I : model specification */
  int           N,             /* I : nb of observation vectors */
  int           D,             /* I : nb of variables */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  int           K,             /* I : number of clusters */
  MissET        Miss,          /* I : how to treat missing data */
  const double*  OldDisp_KD,    /* I : old dispersions (K,D,D) */

  double*        OldNewCen_KD,  /* I/O : old then updated means (K,D) */

  int*          EmptyK_P,      /* O : index of empty class (0 or 1 ..K) */
  double*        N_K,           /* O : size of a class (K) */
  double*        N_KD,          /* O : size of a class and variable (K,D) */
  double*        N_KDD,          /* O : size of a class and variable (K,D,D) */
  double*        Iner_KD        /* O : inertia of a class/variable (K,D,D) */
 ) ; 


static void InerToDisp
(
 DisperET      DispType, /* I : dispersion model */
 int           N,        /* I : nb of observation vectors */
 int           Nk,       /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const ModelSpecT* ModelSpecP,
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  NbObs_KDD, /* I : nb of observations in each 
                             class/variable/variable */
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD,  /* O : dispersion in each class/variable */
 StatusET*     StsP      /* [O] : STS_E_FUNCARG if unknown DispType */
) ;



/* Called by InerToDisp() */

static void InerToDisp__
(
 const ModelSpecT* ModelSpecP,
 int           N,        /* I : nb of observation vectors */
 int           Nk,       /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD   /* O : dispersion in each class/variable */
) ;

static void InerToDispK_
(
 const ModelSpecT* ModelSpecP,
 int           N,        /* I : nb of observation vectors */
 int           K,        /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD   /* O : dispersion in each class/variable */
) ;


static void InerToDisp_D
(
 const ModelSpecT* ModelSpecP,
 int           N,        /* I : nb of observation vectors */
 int           K,        /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  NbObs_KDD, /* I : nb of observations in each class/variable/variable*/
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD   /* O : dispersion in each class/variable */
) ;


static void InerToDispKD
(
 const ModelSpecT* ModelSpecP,
 int           N,        /* I : nb of observation vectors */
 int           K,        /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  NbObs_KDD, /* I : nb of observations in each class/variable/variable*/
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD   /* O : dispersion in each class/variable */
) ;



/* Called by EstimParaLaplace() */


/*V1.06-f*/
static StatusET          /* Return status : OK, EMPTY or MEMORY */
CommonLaplaceDiag
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const int*    Sort_ND,       /* I : sorted index / variable (N,D) */
  int           N,             /* I : nb of observation vectors */
  int           D,             /* I : nb of variables */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  int           K,             /* I : number of clusters */
  MissET        Miss,          /* I : how to treat missing data */
  const double*  OldDisp_KD,    /* I : old dispersions (K,D) */

  double*        OldNewCen_KD,  /* I/O : old then updated means (K,D) */

  int*          EmptyK_P,      /* O : index of empty class (0 or 1 ..K) */
  double*        N_K,           /* O : size of a class (K) */
  double*        N_KD,          /* O : size of a class and variable (K,D) */
  double*        Iner_KD        /* O : inertia of a class/variable (K,D) */
 ) ; 


#if 0 /* already declared above */

static void InerToDisp
(
 DisperET      DispType, /* I : dispersion model */
 int           N,        /* I : nb of observation vectors */
 int           Nk,       /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD   /* O : dispersion in each class/variable */
) ;

#endif


/* Called by CommonLaplaceDiag() */

static void EstimSizes
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  const int     N,             /* I : nb of observation vectors */
  const int     D,             /* I : nb of variables */
  const int     K,             /* I : number of clusters */
  double*        N_K,           /* O : size of a class (K) */
  double*        N_KD           /* O : size of a class and variable (K,D) */  
 ) ;


static StatusET          /* Return status : OK, EMPTY */
EstimLaplaceCenters
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const int*    Sort_ND,       /* I : sorted index / variable (N,D) */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  const int     N,             /* I : nb of observation vectors */
  const int     D,             /* I : nb of variables */
  const int     K,             /* I : number of clusters */
  const MissET  Miss,          /* I : how to treat missing data */
  const double*  N_K,           /* I : size of a class (K) */
  const double*  N_KD,          /* I : size of a class and variable (K,D) */  
  const double*  OldCen_KD,     /* I : old centers */
  const double*  OldDisp_KD,    /* I : old dispersions (K,D) */

  double*        OldNewCen_KD,  /* I/O : eventually updated centers (K,D) */
  int*          EmptyK_P       /* O : index of empty class (0 or 1 ..K) */
 ) ;


static void EstimLaplaceIner
 ( 
  const double*  X_ND,          /* I : data matrix (N,D) */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  const int     N,             /* I : nb of observation vectors */
  const int     D,             /* I : nb of variables */
  const int     K,             /* I : number of clusters */
  const MissET  Miss,          /* I : how to treat missing data */
  const double*  N_K,           /* I : size of a class (K) */
  const double*  N_KD,          /* I : size of a class and variable (K,D) */  
  const double*  OldCen_KD,     /* I : old centers */
  const double*  OldDisp_KD,    /* I : old dispersions (K,D) */
  const double*  NewCen_KD,     /* I : new centers */
  double*        Iner_KD        /* O : inertia of a class/variable (K,D) */  
 ) ;



/* Called by EstimLaplaceCenters() */

static void ComputeMedian
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const int*    Sort_ND,       /* I : sorted index / variable (N,D) */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  const int     N,             /* I : nb of observation vectors */
  const int     D,             /* I : nb of variables */
  const int     K,             /* I : number of clusters */
  const int     H,             /* I : current cluster */
  const int     J,             /* I : current variable */
  const double   totwei,        /* I : total weight */
  int*          ImedP,         /* O : index of median of observed values */
  double*        CumweiP,       /* O : cumulated weights until median obs. */
  double*        MedvalP        /* O : median value (eventually midway) */
 ) ;



static double FindMinInerLaplaceEM  /* ret: minimizer of expected inertium */
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const int*    Sort_ND,       /* I : sorted index / variable (N,D) */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  const int     N,             /* I : nb of observation vectors */
  const int     D,             /* I : nb of variables */
  const int     K,             /* I : number of clusters */
  const int     H,             /* I : current cluster */
  const int     J,             /* I : current variable */
  const int     Imed,          /* I : median position of observed data */
  const double   Cumwei,        /* I : cumulated weight until Xmed */
  const double   Cen0,          /* I : previous center */
  const double   Disp0,         /* I : previous dispersion */
  const double   Nhj,           /* I : number of observations */
  const double   Nmis           /* I : number of missing data */
 ) ;



/* Called by FindMinInerLaplaceEM() */

static double DerivInerDir      /* ret: +/- R'(Y), R expected inertia */ 
 (
  const double   Y,             /* I : point at which to compute R' */
  const double   Weidif,        /* I : "after+med-bef" or "bef+med-after" */
  const double   Intwei,        /* I : cumulated intermediate weight */
  const double   Nmis,          /* I : number of missing data */
  const double   Cen0,          /* I : previous center */
  const double   Disp0          /* I : previous dispersion */
 ) ;




/* ==================== GLOBAL FUNCTION DEFINITION =================== */


/* ------------------------------------------------------------------- */
int GetDensityFunc  /* STS_OK or STS_E_FUNCARG */
        (
            const ModelSpecT  *SpecP,           /* I */
            CompuDensFT**     CompuDensFP       /* O */
        )
/* ------------------------------------------------------------------- */
{

    switch( SpecP->ClassFamily )
    {
        case FAMILY_NORMALGEN:
        case FAMILY_NORMALDIAG:
	  *CompuDensFP = DensNormal ;
	  return STS_OK ;

        case FAMILY_LAPLACE:
	  *CompuDensFP = DensLaplaceDiag ;
	  return STS_OK ;

        default :
	  *CompuDensFP = NULL ;
	  fprintf( stderr, "GetDensityFunc bad arg : family = %d\n",
                   SpecP->ClassFamily ) ;
	  return STS_E_FUNCARG ;
    }

  /*???*/
}   /* end of GetDensityFunc() */




/* ------------------------------------------------------------------- */
StatusET            /* ret : OK, W_EMPTYCLASS or E_MEMORY */ /*V1.06-b*/
EstimPara 
(
  const double      *C_NK,      /* I : classification matrix (N,K) */
  const DataT      *DataP,     /* I : observed points */
  const AlgoParaT  *AlgoParaP,
  int               Nk,         /* I : number of classes */
  MissET            MissMode,   /* I : how to treat missing data */ /*V1.05-b*/
  const ModelSpecT * SpecP,  /* I : model specification */ /*V1.06-b*/

  int              *EmptyK_P,  /* O : which empty class (1..K) or 0 */ /*V1.05-b*/
  ModelParaT       *ParaP      /* I/O : previous and new estimated parameters */
)
/* ------------------------------------------------------------------- */
{
  StatusET    sts ;  /* return status */

  int         k ;
  
  /* Family dependent estimation method 
   */
  switch( SpecP->ClassFamily ) {
  case FAMILY_NORMALGEN:
  case FAMILY_NORMALDIAG:
    sts = EstimParaNormal( C_NK, DataP, Nk, MissMode, SpecP, 
			   EmptyK_P, ParaP ) ;


    break ;

  case FAMILY_LAPLACE:
    sts = EstimParaLaplace( C_NK, DataP, Nk, MissMode, SpecP, 
			    EmptyK_P, ParaP ) ;      
    break;

  default:
    sts = STS_E_FUNCARG ;
  }
  
  
 switch ( AlgoParaP->Algo) 
   {
    case ALGO_NEM:
    case ALGO_NCEM:
       /* Estimate proportions */
       if ( SpecP->ClassPropor == PROPOR_K ) {
	 for ( k = 0; k < Nk ; k ++ )
	   ParaP->Prop_K[ k ] = ParaP->NbObs_K[ k ] / DataP->NbPts ;
       }
       else {
	for ( k = 0; k < Nk ; k ++ )
	  ParaP->Prop_K[ k ] = 1.0 / Nk ;
       }
       break;

   default: break;
     /* case of MRF, no proportion to estimate */
   
   }

  return sts ;
  /*???*/
}   /* end of EstimPara() */


/* ------------------------------------------------------------------- */
int choldc
(
      double*       a, 
      int          n, 
      double*       p
)
/* given a positive-definite symmetric matrix, a, n by n, this routine construct its cholesky
   decomposition, a=l.l^t. on input, only  the upper triangle of a need by given, it is not 
   modified. l is retourned in the lower triangle of a, except for its diagonal which is 
   returned in p[1,...,n] */
/* ------------------------------------------------------------------- */
{
	int    i,j,k;
	double sum;

	for (i=0;i<n;i++) 
	  {
		for (j=i;j<n;j++) 
		  {
			for (sum=a[i*n +j],k=i-1;k>=0;k--) sum -= a[i*n +k]*a[j*n +k];
			if (i == j) 
                        {
			  if (sum <= 0.0)
			    {
			    return -1; /*choldc failed */;                                       
			    fprintf(stderr,"cholesky failed\n");
			    }             
			   p[i]=sqrt(sum);
			} 
                        else a[j*n +i]=sum/p[i];
		}
	}
	return 0;
}
/* (C) Copr. 1986-92 Numerical Recipes Software 7L`2s5N:Z. */
/* ------------------------------------------------------------------- */



/* ==================== LOCAL FUNCTION DEFINITION =================== */





/* ------------------------------------------------------------------- */
static int DensNormal      /* ret : 0 if OK, -1 if zero density */
        (
            int                Nd,         /* I : point dimension */
            int                Ik,         /* I : class number : 0..Nk-1 */
            const ModelParaT*  ParaP ,     /* I : noise parameters */
            const double*       XV,         /* I : point (dim d) */
            const ModelSpecT*  ModelSpecP,
            double*            FkP,        /* O : density for class Ik */
            double*             LogFkP,      /* O : log of density */
            const double*            InvChol, /* inverse of cholesky decomp L of the cov. matrix*/
            const double            logdet 
        )
/* ------------------------------------------------------------------- */
{ 
  int     err=0;
  int     d,i,j;      /* 0..Nd-1 : current variable */
  double   dk ;     /* sum_d [ log(2pi skd2) + (xd-mkd)^2/skd2 ] */
  int     nbobs ;  /* 0..Nd-1 : nb of observed variables */ /*V1.05-b*/
  int     nuldisp=0; /* TRUE if one of the dispersions is zero */
  double  sum = 0;
 

  switch ( ModelSpecP->ClassFamily)
    {
    case FAMILY_NORMALDIAG:
    /* Gaussian mixture density, diagonal models :

     Complete data :
     fk(x) = [ prod_d (2*pi * vkd)^(-0.5) ] * exp( -0.5 * dist_x_mk )
     where  dist_x_mk = sum[d=1,Nd]( (x(d) - mk(d))^2 ) / vkd

     Incomplete data :
     fk(x) = [ prod_{d in oi} (2*pi * vkd)^(-0.5) ] * exp( -0.5 * dist_x_mk )
     where   dist_x_mk = sum[d in oi]  (x(d) - mk(d))^2 / vk
     =>
     log fk(x) = - 0.5 sum_{d in oi}[ log(2*pi * vkd) + (xd-mkd)^2/vkd ]

  */

  for ( d = 0, 
          dk = 0.0, nbobs = 0, nuldisp = 0 ; 
        d < Nd ; 
        d ++ )
    {
      if ( ! isnan( XV[ d ] ) )   /*V1.05-b*/
        {
          double dif = XV[ d ] - ParaP->Center_KD[ (Ik * Nd) + d ] ;

          if ( ParaP->Disp_KD[Ik*Nd + d ] > EPSILON )
            dk = dk 
	      + log( TWO_PI * ParaP->Disp_KD[Ik*Nd + d  ] ) 
	      + (dif * dif) / ParaP->Disp_KD[Ik*Nd + d  ];
         else
             nuldisp = 1;

	  nbobs++;
	}
    }

    if ( ! nuldisp )
    {
      /*V1.04-a*//*V1.05-b*/
      *LogFkP = -0.5 * dk ;
      *FkP = exp( *LogFkP ) ;
      return 0 ;
    }
   else
    {

      if ( XV[ d ] != ParaP->Center_KD[ (Ik * Nd) + d ])
        {
         *LogFkP = - MAXDOUBLE ;
         *FkP = 0.0 ;
         return -1 ;
	}

      if ( XV[ d ] == ParaP->Center_KD[ (Ik * Nd) + d ])
	{
         *LogFkP = MAXDOUBLE;
         *FkP = MAXDOUBLE;
         return 0 ;
	}

    }
     

    case FAMILY_NORMALGEN: /* only for complete data  */
   
      /* compute log(f(Yi|class = Ik) */ 
       dk = 0;
          /* compute L^-1.(Yi-meanIk) */
          for (i=0; i<Nd; i++)
          {
	    sum = 0;
	    for (j=0; j<i+1; j++)
	    {
	     sum += InvChol[i*Nd+j]*(XV[j] - ParaP->Center_KD[Ik*Nd+j]);
	    }
	    dk = dk +  sqr(sum)+ log(TWO_PI);
           }
       dk +=logdet;

      *LogFkP = -0.5 * dk ;
      *FkP = exp( *LogFkP ) ;
      return 0 ;


    break;
   
    case FAMILY_LAPLACE:
    case FAMILY_NB:
    default:
      err = -1;
      return err;

   }/* end of switch */

}  /* end of DensNormal() */



/* ------------------------------------------------------------------- */
static int DensLaplaceDiag      /* ret : 0 if OK, -1 if zero density */
        (
            int                Nd,         /* I : point dimension */
            int                Ik,         /* I : class number : 0..Nk-1 */
            const ModelParaT*  ParaP ,     /* I : noise parameters */
            const double*       XV,         /* I : point (dim d) */
            const ModelSpecT*  ModelSpecP,
            double*            FkP,        /* O : density for class Ik */
            double*            LogFkP,      /* O : log of density */
            const double*            InvChol,
            const double            logdet 
        )
/* ------------------------------------------------------------------- */
{
  int     d ;      /* 0..Nd-1 : current variable */
  double   dk ;     /* sum_d [ log(2 lkd) + |xd-mkd| / lkd ] */
  int     nbobs ;  /* 0..Nd-1 : nb of observed variables */ /*V1.05-b*/
  int     nuldisp; /* TRUE if one of the dispersions is zero */

  /* Laplace mixture density, diagonal models :

     Complete data :
     fk(x) = [ prod_d (2*vkd)^(-1) ] * exp( - dist_x_mk )
     where  dist_x_mk = sum[d=1,Nd] | x(d) - mk(d) | / vkd

     Incomplete data :
     fk(x) = [ prod_{d in oi} (2*vkd)^(-1) ] * exp( - dist_x_mk )
     where   dist_x_mk = sum[d in oi]  | x(d) - mk(d) | / vkd
     =>
     log fk(x) = - sum_{d in oi}[ log(2*vkd) + |xd-mkd| / vkd ]

  */

  for ( d = 0, 
          dk = 0.0, nbobs = 0, nuldisp = 0 ; 
        d < Nd ; 
        d ++ )
    {
      if ( ! isnan( XV[ d ] ) )   /*V1.05-b*/
        {
          double dif = XV[ d ] - ParaP->Center_KD[ (Ik * Nd) + d ] ;

          if ( ParaP->Disp_KD[Ik*Nd*(Nd+1)/2 +d*(2*Nd+1-d)/2   ] > EPSILON )
            dk = dk 
      	+ log( 2 * ParaP->Disp_KD[ Ik*Nd*(Nd+1)/2 +d*(2*Nd+1-d)/2 ] ) 
      	+ fabs( dif ) / ParaP->Disp_KD[Ik*Nd*(Nd+1)/2 +d*(2*Nd+1-d)/2  ] ;
          else
            nuldisp = 1 ;

          nbobs ++ ;
        }
  }

  if ( ! nuldisp )
    {
      /*V1.04-a*//*V1.05-b*/
      *LogFkP = - dk ;
      *FkP = exp( *LogFkP ) ;
      return 0 ;
    }
  else
    {
      *LogFkP = - MAXDOUBLE ;
      *FkP = 0.0 ;
      return -1 ;
    }

}   /* end of DensLaplaceDiag() */




/* ------------------------------------------------------------------- */
static StatusET            /* ret : OK, W_EMPTYCLASS or E_MEMORY */ /*V1.06-b*/
EstimParaNormal 
(
  const double   *C_NK,      /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           Nk,         /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/
  const ModelSpecT* SpecP,  /* I : model specification */ /*V1.06-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 */ /*V1.05-b*/
  ModelParaT    *ParaP      /* O : estimated parameters */
)
/* ------------------------------------------------------------------- */
{
  int      N = DataP->NbPts ;
  int      D = DataP->NbVars ;

  StatusET sts ;


  /* Common computation for any  gaussian model with missing data */
  sts = CommonGauss
    ( DataP->PointsM,SpecP, N, D, C_NK, Nk, MissMode, ParaP->Disp_KD, 
      ParaP->Center_KD, 
      EmptyK_P, ParaP->NbObs_K, ParaP->NbObs_KD,  ParaP->NbObs_KDD,ParaP->Iner_KD ) ;

  /* Dispersion is computed depending on dispersion model */
  InerToDisp( SpecP->ClassDisper, N, Nk, D, SpecP, ParaP->NbObs_K, ParaP->NbObs_KD, 
	      ParaP->NbObs_KDD, ParaP->Iner_KD, MissMode, 
	      ParaP->Disp_KD, & sts ) ;
  return sts ;

}   /* end of EstimParaNormal() */



/* ------------------------------------------------------------------- */
static StatusET          /* Return status : OK, EMPTY or MEMORY */
CommonGauss
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const ModelSpecT* ModelSpecP,
  int           N,             /* I : nb of observation vectors */
  int           D,             /* I : nb of variables */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  int           K,             /* I : number of clusters */
  MissET        Miss,          /* I : how to treat missing data */
  const double* OldDisp_KD,    /* I : old dispersions (K,D,D) */

  double*        OldNewCen_KD,  /* I/O : old then updated means (K,D) */

  int*          EmptyK_P,      /* O : index of empty class (0 or 1 ..K) */
  double*        N_K,           /* O : size of a class (K) */
  double*        N_KD,          /* O : size of a class and variable (K,D) */
  double*        N_KDD,          /* O : size of a class and variable (K,D,D) */
  double*       Iner_KD        /* O : inertia of a class/variable (K,D,D) */
 ) 
/*\

   This function does the common computation that is needed at the
   M-step, in all the Gaussian models (with missing data traited if 
   diagonal model).

   The routine takes as input a data matrix Xij_nd (possibly
   containing NaN values), a classification matrix Cih_nk, and the
   kind of missing data processing Miss (REPLACE or IGNORE) ; and also
   the old volumes OldVh_k, covariance matrices OldChjj_kdd, and means
   OldNewMhj_kd of the classes, to be used in the REPLACE mode.

   It produces as output the reestimated mean vectors in OldNewMhj_kd;
   reestimated proportions in NewP_k; the last encountered empty class
   in EmptyK_P (or 0 if no empty class); the {sum[i] cih} in Nh_k, the
   {sum[i] cih rij} in Nhj_kd (number of present observations in class h
   and variable j); and the "missing data dispersion" of each
   class/variable in Disphj_kd (i.e. the missing data equivalent of
   {sum[i] Cih (xij - mhj)^2}).

\*/
/* ------------------------------------------------------------------- */
{
  const char* func = "CommonGauss" ;

  double*   sumdata_KD ;    /* sum[i] cih rij xij   (K,D), local alloc */
  double*   sumsquare_KD ;  /* sum[i] cih rij xij^2 (K,D), local alloc */
  double*   oldmean_KD ;    /* saved previous means (K,D), local alloc */

  int      h ;             /* current class    0..K-1 */
  int      j ;             /* current variable 0..D-1 */
  int      i,s ;          
  StatusET sts ;           /* return status */

  double    obsinerhj ;
  double    numcovhij ;
  double    ysi ;
  double    ysj ;
  double    csh ;

  sts = STS_OK ;

  /* Allocate local structures */
  sumdata_KD   = GenAlloc( K * D, sizeof(double), 1, func, "sumdata" ) ;
  sumsquare_KD = GenAlloc( K * D, sizeof(double), 1, func, "sumsquare" );
  oldmean_KD   = GenAlloc( K * D, sizeof(double), 1, func, "oldmean" ) ;
  
  if ( (sumdata_KD == NULL) || (sumsquare_KD == NULL) || (oldmean_KD== NULL) )
    fprintf(stderr," impossible to allocate variables in function CommonGauss\n"); 

  /* Save old mean value */
  memcpy( oldmean_KD , OldNewCen_KD , K * D * sizeof( double ) ) ;

  /* Set no empty class by default */
  (*EmptyK_P) = 0 ;

  if (ModelSpecP->ClassFamily ==FAMILY_NORMALDIAG)
  {
  /* For each class h */
  for ( h = 0 ; h < K ; h ++ ) {
  
    /* compute the new means and the numerator Iner_KD of the new variances */
      /* For each variable j */
      for ( j = 0 ; j < D ; j ++ ) {

	  /* Initialize the sum[i] quantities to 0 */
	  N_K[ h ]             = 0.0 ;
	  N_KD[ _HJ ]          = 0.0 ;
	  sumdata_KD[ _HJ ]    = 0.0 ;
	  sumsquare_KD[ _HJ ]  = 0.0 ;

	  /* For each object i */
	  for ( i = 0 ; i < N ; i ++ ) {

	      double cih = C_NK[ _IH ] ;
	      double xij = X_ND[ _IJ ] ;

	      /* Increment class size */
	      N_K[ h ] += cih ;

	      /* If this object's j variable is observed */
	      if ( ! isnan( xij ) )
		{
		  /* Increment all sum[i] of observed quantities */
		  N_KD[ _HJ ]          += cih ;
		  sumdata_KD[ _HJ ]    += cih * xij ;
		  sumsquare_KD[ _HJ ]  += cih * xij * xij ;
		}
	      /* Else xij is missing, don't change sum[i] of observed data */
	  }
	  /* End  For each object i (all sums[i] are now done) */

	  /* If this class size > 0 */
	  if ( N_K[ h ] > 0 ) {
	      /* 
	       * Compute means Mhj and dispersions Dhj from computed sums,
	       * depending on REPLACE or IGNORE missing mode 
	       */
	      if ( Miss == MISSING_REPLACE ) {
		  /* 
		   * new_mhj = 
		   *  ( sum[i] cih rij xij + ( nh - nhj ) old_mhj ) / nh 
		   *
		   * disp_hj = 
		   *  "observed dispersion"_hj +
		   *  ( nh - nhj ) [ (new_mhj - old_mhj)^2 + old_vh old_chjj ]
		   */

		  OldNewCen_KD[ _HJ ] = 
		    ( sumdata_KD[ _HJ ] + 
		      ( N_K[ h ] - N_KD[ _HJ ] ) * oldmean_KD[ _HJ ] )
		    / N_K[ h ] ;

		  /* V1.06-d
		   * "observed dispersion"_hj = 
		   *      sum[i] cih rij (xij - new_mhj)^2 =
		   *      sum[i] cih rij xij^2 - 
		   *       new_mhj * ( 2 * sum[i] cih rij xij - nhj new_mhj )
		   */
		  obsinerhj = sumsquare_KD[ _HJ ] - 
		    OldNewCen_KD[ _HJ ] * 
		    (2*sumdata_KD[ _HJ ] - N_KD[ _HJ ] * OldNewCen_KD[ _HJ ]) ;

		  Iner_KD[h*D + j] = 
		    obsinerhj +
		    ( N_K[ h ] - N_KD[ _HJ ] ) *
		    ( sqr( OldNewCen_KD[ _HJ ] - oldmean_KD[ _HJ ] ) +
		      OldDisp_KD[ h*D + j] ) ;
	      }
	      else  { /* assume Miss == MISSING_IGNORE */

		  /* 
		   * new_mhj = 
		   *  . if  nhj <> 0,  ( sum[i] cih rij xij ) / nhj  
		   *  . if  nhj == 0,  old_mhj
		   *
		   * disp_hj = 
		   *  "observed dispersion"_hj 
		   */

		  if ( N_KD[ _HJ ] > 0 )
		    OldNewCen_KD[ _HJ ] = sumdata_KD[ _HJ ] / N_KD[ _HJ ];
		  else
		    OldNewCen_KD[ _HJ ] = oldmean_KD[ _HJ ] ;

		  /*
		   * "observed dispersion"_hj = 
		   *      sum[i] cih rij (xij - new_mhj)^2 =
		   *      sum[i] cih rij xij^2 - nhj new_mhj^2
		   */
		  obsinerhj = sumsquare_KD[ _HJ ] - 
		    N_KD[ _HJ ] * sqr( OldNewCen_KD[ _HJ ] ) ;

		  Iner_KD[h*D + j] = obsinerhj ;
	      }

	  }
	  else /* then this class size == 0 */ {
	      /* signal empty class and which one */
	      sts = STS_W_EMPTYCLASS ;
	      (*EmptyK_P) = h + 1 ;
	  }

       }/* end  For each variable j */
    } /* end  For each class h */
  }

                             /*   ~~~~~~~~~~~~~~~~~~~~~~ */

 
  if (ModelSpecP->ClassFamily ==FAMILY_NORMALGEN)/* by default, missing-ignore */
  {
    switch (ModelSpecP->ClassDisper) {
    case DISPER__D: 

      /* compute the new means and the numerator Iner_KD of the new variances */
      for (h=0; h<K;h++)             /* first class    */
      {
      /* For each variable j */
      for ( j = 0 ; j < D ; j ++ ) {

	  /* Initialize the sum[i] quantities to 0 */
	  N_K[ h ]             = 0.0 ;
	  N_KD[ _HJ ]          = 0.0 ;
	  sumdata_KD[ _HJ ]    = 0.0 ;
	  sumsquare_KD[ _HJ ]  = 0.0 ;

	  /* For each object i */
	  for ( i = 0 ; i < N ; i ++ ) {

	      double cih = C_NK[ _IH ] ;
	      double xij = X_ND[ _IJ ] ;

	      /* Increment class size */
	      N_K[ h ] += cih ;

	      /* If this object's j variable is observed */
	      if ( ! isnan( xij ) )
		{
		  /* Increment all sum[i] of observed quantities */
		  N_KD[ _HJ ]          += cih ;
		  sumdata_KD[ _HJ ]    += cih * xij ;
		  sumsquare_KD[ _HJ ]  += cih * xij * xij ;
	
		}
	      /* Else xij is missing, don't change sum[i] of observed data */
	  }
	  /* End  For each object i (all sums[i] are now done) */
      } /* end  For each variable j */
      } /* end  For each class h */

  for ( j = 0 ; j < D ; j ++ ) {
    obsinerhj = 0 ;
    for (h=0; h<K;h++){
       if ( N_K[ h ] > 0 ) {
	    if ( Miss == MISSING_REPLACE ) { sts = STS_E_FUNCARG; }
	    if ( Miss == MISSING_IGNORE ) {
		  if ( N_KD[ _HJ ] > 0 )
		    { OldNewCen_KD[ _HJ ] = sumdata_KD[ _HJ ] / N_KD[ _HJ ];}
		  else
		    { OldNewCen_KD[ _HJ ] = oldmean_KD[ _HJ ] ;}

		  obsinerhj = obsinerhj + sumsquare_KD[ _HJ ] - N_KD[ _HJ ] * sqr( OldNewCen_KD[ _HJ ] ) ;
	    }
       }
       else /* then this class size == 0 */ {
	      /* signal empty class and which one */
	      sts = STS_W_EMPTYCLASS ;
	      (*EmptyK_P) = h + 1 ;
       }
    }
    Iner_KD[j*(2*D+1-j)/2] = obsinerhj ;	      
  }

     /* compute the numerator of the Eij */
    for (i=0; i<D;i++)             /* first variable */
    {        
        for (j=i+1;j<D;j++)        /* second variable */
	{       
	  numcovhij = 0;
          N_KDD[i*D+j-i*(i+1)/2] = 0;
          for (h = 0; h<K ; h++)
	  {
            if ( N_K[ h ] > 0 ) {
	     for (s=0;s<N;s++)        /* for each site */
	     {
              ysi = X_ND[s*D+i];
              ysj = X_ND[s*D+j];
	      csh = C_NK[s*K+h];
              if ( (!isnan(ysi)) && (!isnan(ysj)) ) 
              {
                N_KDD[i*D+j-i*(i+1)/2] +=csh;
	        numcovhij += (ysi-OldNewCen_KD[h*D+i])*(ysj-OldNewCen_KD[h*D+j])*csh; 
	      }
	     }
	    }
            else /* then this class size == 0 */ 
            {
	      /* signal empty class and which one */
	      sts = STS_W_EMPTYCLASS ;
	      (*EmptyK_P) = h + 1 ;
             }

	  }
	     Iner_KD[i*D+j-i*(i+1)/2]=numcovhij ;
	}
    }    
      break;
   
    case DISPER_KD:
     for (h=0; h<K;h++)             /* first class    */
      {

      /* compute the new means and the numerator Iner_KD of the new variances */
      /* For each variable j */
      for ( j = 0 ; j < D ; j ++ ) {

	  /* Initialize the sum[i] quantities to 0 */
	  N_K[ h ]             = 0.0 ;
	  N_KD[ _HJ ]          = 0.0 ;
	  sumdata_KD[ _HJ ]    = 0.0 ;
	  sumsquare_KD[ _HJ ]  = 0.0 ;

	  /* For each object i */
	  for ( i = 0 ; i < N ; i ++ ) {

	      double cih = C_NK[ _IH ] ;
	      double xij = X_ND[ _IJ ] ;

	      /* Increment class size */
	      N_K[ h ] += cih ;

	      /* If this object's j variable is observed */
	      if ( ! isnan( xij ) )
		{
		  /* Increment all sum[i] of observed quantities */
		  N_KD[ _HJ ]          += cih ;
		  sumdata_KD[ _HJ ]    += cih * xij ;
		  sumsquare_KD[ _HJ ]  += cih * xij * xij ;
		}
	      /* Else xij is missing, don't change sum[i] of observed data */
	  }
	  /* End  For each object i (all sums[i] are now done) */

	  /* If this class size > 0 */
	  if ( N_K[ h ] > 0 ) {
	    if ( Miss == MISSING_REPLACE ) { sts = STS_E_FUNCARG; }
	    if ( Miss == MISSING_IGNORE ) {
		  if ( N_KD[ _HJ ] > 0 )
		    OldNewCen_KD[ _HJ ] = sumdata_KD[ _HJ ] / N_KD[ _HJ ];
		  else
		    OldNewCen_KD[ _HJ ] = oldmean_KD[ _HJ ] ;

		  obsinerhj = sumsquare_KD[ _HJ ] - 
		    N_KD[ _HJ ] * sqr( OldNewCen_KD[ _HJ ] ) ;

		  Iner_KD[h*D*(D+1)/2 + j*(2*D+1-j)/2] = obsinerhj ;
	      }

	  }
	  else /* then this class size == 0 */ {
	      /* signal empty class and which one */
	      sts = STS_W_EMPTYCLASS ;
	      (*EmptyK_P) = h + 1 ;
	  }

       }/* end  For each variable j */


     /* compute the numerator of the Eij */        
    /* If this class size > 0 */
    if ( N_K[ h ] > 0 ) {

    for (i=0; i<D;i++)             /* first variable */
    {        
        for (j=i+1;j<D;j++)        /* second variable */
	{       
	  numcovhij = 0;
          N_KDD[h*D*(D+1)/2 +i*D+j-i*(i+1)/2] = 0;
	  for (s=0;s<N;s++)        /* for each site */
	    {
              ysi = X_ND[s*D+i];
              ysj = X_ND[s*D+j];
	      csh = C_NK[s*K+h];
              if ( (!isnan(ysi)) && (!isnan(ysj)) ) 
              {
                N_KDD[h*D*(D+1)/2 +i*D+j-i*(i+1)/2] +=csh;
	        numcovhij += (ysi-OldNewCen_KD[h*D+i])*(ysj-OldNewCen_KD[h*D+j])*csh; 
	      }
	    }
	     Iner_KD[h*D*(D+1)/2 +i*D+j-i*(i+1)/2]=numcovhij ;
	 }
      }
    }

   else /* then this class size == 0 */ 
   {
	      /* signal empty class and which one */
	      sts = STS_W_EMPTYCLASS ;
	      (*EmptyK_P) = h + 1 ;
   }
  }/* end for each class h */
     break;

    default:
      sts = STS_E_FUNCARG;
      
   } /* end switch DisperType  */
  } /* end NORMALGEN */


  /* Free local structures */
  GenFree( oldmean_KD ) ;
  GenFree( sumsquare_KD ) ;
  GenFree( sumdata_KD  ) ;

  return sts ;

}   /* end of CommonGauss() */



/* ------------------------------------------------------------------- */
static void InerToDisp
(
 DisperET      DispType, /* I : dispersion model */
 int           N,        /* I : nb of observation vectors */
 int           Nk,       /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const ModelSpecT* ModelSpecP,
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  NbObs_KDD, /* I : nb of observations in each class/variable*/
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD,  /* O : dispersion in each class/variable */
 StatusET*     StsP      /* [O] : STS_E_FUNCARG if unknown DispType */
)
/* ------------------------------------------------------------------- */
{

  if (ModelSpecP->ClassFamily == FAMILY_NORMALDIAG)
    { 
  switch( DispType ) {
  case DISPER___:  /* Common dispersion for all classes and variables, 
                                                           Ek=s^2Id */
    InerToDisp__(  ModelSpecP,N, Nk, D, NbObs_K, NbObs_KD,Iner_KD, MissMode, 
		  Disp_KD ) ;
    break ;

  case DISPER_K_:  /* In each class, common dispersion for all variables,
                                                                 Ek=sk^2Id */
    InerToDispK_( ModelSpecP,N, Nk, D, NbObs_K, NbObs_KD, Iner_KD, MissMode, 
		  Disp_KD ) ;
    break ;

  case DISPER__D:  /* In each variable, common dispersion for all classes,
                                Ek=E diagonal matrix with E(i,i) not= E(j,j) */
    InerToDisp_D( ModelSpecP,N, Nk, D, NbObs_K, NbObs_KD, NbObs_KDD, Iner_KD, MissMode, 
		  Disp_KD ) ;
    break ;

  case DISPER_KD:  /* In each class, common dispersion for all variables
                                Ek diagonal matrix with Ek(i,i) not= Ek(j,j)  */
    InerToDispKD( ModelSpecP,N, Nk, D, NbObs_K, NbObs_KD, NbObs_KDD, Iner_KD, MissMode, 
		  Disp_KD ) ;
    break ;

  default:
    (*StsP) = STS_E_FUNCARG ;
  }
    }

  if (ModelSpecP->ClassFamily == FAMILY_NORMALGEN)
    { 
  switch( DispType ) {
  case DISPER__D:  /* Common dispersion for all classes and variables,
                                Ek=E, full matrix */
    InerToDisp_D( ModelSpecP,N, Nk, D, NbObs_K, NbObs_KD, NbObs_KDD, Iner_KD, MissMode, 
		  Disp_KD ) ;
    break ;

  case DISPER_KD:  /* In each class, common dispersion for all variables,
                                Ek full matrix */
    InerToDispKD( ModelSpecP,N, Nk, D, NbObs_K, NbObs_KD, NbObs_KDD, Iner_KD, MissMode, 
		  Disp_KD ) ;
    break ;

  default:
    (*StsP) = STS_E_FUNCARG ;
  }
    }

}   /* end of InerToDisp() */


/* ------------------------------------------------------------------- */
static void InerToDisp__
(
 const ModelSpecT* ModelSpecP,
 int           N,        /* I : nb of observation vectors */
 int           Nk,       /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD   /* O : dispersion in each class/variable */
)
/* ------------------------------------------------------------------- */
{
  int      k ;         /* current class    0..K-1 */
  int      d;         /* current variable 0..D-1 */

  double    vol ;       /* common volume */
  double    nobs ;      /* number of effective observations : ]0,N*D] */

  /* Compute common volume : 
     sum dispersions over the classes and variables 
     then divide by all/observed data size */
  
  vol = 0.0 ;
  nobs = 0.0 ;


 if (ModelSpecP->ClassFamily == FAMILY_NORMALDIAG)
  { 

  for ( k = 0; k < Nk ; k ++ )
    {

      /* If not empty class */
      if ( NbObs_K[ k ] > 0 )
	{
	  for ( d = 0 ; d < D ; d ++ )
	    {
	      vol  += Iner_KD[k*D + d] ;
	      nobs += NbObs_KD[k*D +d] ; 
	    }
	}
    }

  if ( MissMode == MISSING_REPLACE )
    vol /= ( N * D ) ;
  else
    vol /= nobs ;  /* nobs must be = N * D - DataP->NbMiss */

  /* For each class k and variable d */
  for ( k = 0 ; k < Nk ; k ++ )
    for ( d = 0 ; d < D ; d ++ )
    {
      /* Set its volume to the common volume */
      Disp_KD[k*D + d ] = vol ;
    }
  }
 
 else /* FAMILY_NORMALGEN*/
   fprintf(stderr,"this DispType is not available for the choosen ClassFamily\n");

}   /* end of InerToDisp__() */


/* ------------------------------------------------------------------- */
static void InerToDispK_
(
 const ModelSpecT* ModelSpecP,
 int           N,        /* I : nb of observation vectors */
 int           K,        /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD   /* O : dispersion in each class/variable */
)
/* ------------------------------------------------------------------- */
{
  int      k ;         /* current class    0..K-1 */
  int      d ;         /* current variable 0..D-1 */

  double    sumd_nkd ;    /* size of obs class */
  double    sumd_inerkd ; /* obs inertia in class k */
  double    dispk ;       /* dispersion in class k */


 if (ModelSpecP->ClassFamily == FAMILY_NORMALDIAG)
  {
  /* For each class k */
  for ( k = 0; k < K ; k ++ ) {

      /* If not empty class */
      if ( NbObs_K[ k ] > 0 ) {

	  /* Reestimate its volume: sum inertia over the variables 
	     then divide by all/observed data size in this class */

	  sumd_nkd    = 0.0 ;
	  sumd_inerkd = 0.0 ;  /* use first component as sum first */

	  for ( d = 0 ; d < D ; d ++ )
	    {
	      sumd_nkd     += NbObs_KD[ k * D + d ] ;
	      sumd_inerkd  += Iner_KD[k*D+ d] ;
	    }

	  if ( MissMode == MISSING_REPLACE )
	    dispk = sumd_inerkd /= ( D * NbObs_K[ k ] ) ;  /*V1.06-e*/
	  else
	    dispk = sumd_inerkd / sumd_nkd ;

	  /* Copy same dispersion into all variables of this class */
	  for ( d = 0 ; d < D ; d ++ )
	    {
	      Disp_KD[k*D + d ] = dispk ;
	    }
      }
      /* (Else class k is empty : keep old dispersion */ 

    }

  /*??*/

  }
 
 else /* FAMILY_NORMALGEN*/
   fprintf(stderr,"this DispType is not available for the choosen ClassFamily\n");
}   /* end of InerToDispK_() */




/* ------------------------------------------------------------------- */
static void InerToDisp_D
(
 const ModelSpecT* ModelSpecP,
 int           N,        /* I : nb of observation vectors */
 int           K,        /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  NbObs_KDD, /* I : nb of observations in each class/variable/variable*/
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD   /* O : dispersion in each class/variable */
)
/* ------------------------------------------------------------------- */
{
  int      k ;         /* current class    0..K-1 */
  int      d,i,j ;         /* current variable 0..D-1 */


 if (ModelSpecP->ClassFamily == FAMILY_NORMALDIAG)
   {
  double    sumk_nkd ;    /* obs size of variable d */
  double    sumk_inerkd ; /* obs inertia in variable d */
  double    dispd ;       /* dispersion of variable d */

  /* For each variable d */
  for ( d = 0; d < D ; d ++ ) {

    /* Reestimate its volume: sum inertia over the classes 
       then divide by all/observed data size in this variable */

    sumk_nkd    = 0.0 ;
    sumk_inerkd = 0.0 ;

    for ( k = 0 ; k < K ; k ++ ) {
      sumk_nkd    += NbObs_KD[ k * D + d ] ;
      sumk_inerkd += Iner_KD[ k*D + d] ;
    }

    if ( MissMode == MISSING_REPLACE )
      dispd = sumk_inerkd / N ;
    else
      dispd = sumk_inerkd / sumk_nkd ;

    /* Copy same dispersion of this variable into all classes */
    for ( k = 0 ; k < K ; k ++ ) {
      Disp_KD[k*D+ d ] = dispd ;
    }
  }
   }

 if (ModelSpecP->ClassFamily == FAMILY_NORMALGEN)
   /* by default missing ignore */
  {
    double nobsi    ;

    /* compute the Eij terms of the covariance matrix for i not= j */
    for (i =0; i<D; i++)
      {
	for (j = i+1; j<D; j++)
	  {
                Disp_KD[i*D+j-i*(i+1)/2]= Iner_KD[i*D+j-i*(i+1)/2]/NbObs_KDD[i*D+j-i*(i+1)/2];
	      }
	      
      }
      

    /* compute the Eii terms of the covariance matrix */   
    for (i =0; i<D; i++)
      {
	nobsi=0;
	for (k=0; k<K; k++)
	   {
	     if (NbObs_K[k]>0)
	       {
                 nobsi+=NbObs_KD[k*D+i];
	       }
	   }
 	    for (k=0; k<K; k++)
	      {
                  Disp_KD[i*(2*D+1-i)/2 ]=Iner_KD[i*(2*D+1-i)/2 ]/nobsi;
	      }
	   
      }


  }

  /*???*/

}   /* end of InerToDisp_D() */



/* ------------------------------------------------------------------- */
static void InerToDispKD
(
 const ModelSpecT* ModelSpecP,
 int           N,        /* I : nb of observation vectors */
 int           K,        /* I : nb of classes */
 int           D,        /* I : nb of variables */
 const double*  NbObs_K,  /* I : nb of observation vectors in each class */
 const double*  NbObs_KD, /* I : nb of observations in each class/variable*/
 const double*  NbObs_KDD, /* I : nb of observations in each class/variable/variable*/
 const double*  Iner_KD,  /* I : inertia of each class/variable */
 MissET        MissMode, /* I : how to treat missing data */ /*V1.05-b*/
 double*        Disp_KD   /* O : dispersion in each class/variable */
)
/* ------------------------------------------------------------------- */
{
  int      k ;         /* current class    0..K-1 */
  int      d,i,j ;         /* current variable 0..D-1 */


 if (ModelSpecP->ClassFamily == FAMILY_NORMALDIAG)
   {
  /* For each class k */
  for ( k = 0 ; k < K ; k ++ ) {

    /* For each variable d */
    for ( d = 0; d < D ; d ++ ) {

      /* Reestimate its volume: divide inertia by all/observed data 
	 size in this variable/ class if not empty */

      if ( MissMode == MISSING_REPLACE ) {
	if ( NbObs_K[ k ] > EPSILON )
	  Disp_KD[k*D + d  ] = 
                   Iner_KD[k*D+ d  ] / NbObs_K[ k ] ;
	/* else empty class : do not update dispersion */
      }
      else /* IGNORE */ {
	if ( NbObs_KD[ k * D + d ] > EPSILON )
	  Disp_KD[ k*D + d ] = 
                 Iner_KD[k*D + d ] /  NbObs_KD[ k * D + d ] ;
      }
     }
    }
   }


 if (ModelSpecP->ClassFamily == FAMILY_NORMALGEN)
   /* by default missing ignore */
  {
  for ( k = 0 ; k < K ; k ++ ) 
    {
    for ( i = 0; i < D ; i ++ ) 
      {
	/* compute the Eii terms of the covariance matrix */
        if ( NbObs_KD[ k * D + i ] > EPSILON )
	  Disp_KD[k*D*(D+1)/2 + i*(2*D+1-i)/2 ] =
             Iner_KD[k*D*(D+1)/2 + i*(2*D+1-i)/2  ] / NbObs_KD[ k * D + i ] ;

        /* compute the Eij terms of the covariance matrix, i not=j */
	for (j=i+1;j<D;j++)
	  {
           if ( NbObs_KDD[k*D*(D+1)/2+i* D + j-i*(i+1)/2 ] > EPSILON )
	     Disp_KD[k*D*(D+1)/2+i* D + j-i*(i+1)/2 ] = 
                      Iner_KD[k*D*(D+1)/2+i* D + j-i*(i+1)/2 ] / 
                                        NbObs_KDD[k*D*(D+1)/2+i* D + j-i*(i+1)/2 ] ;
	  }
      }
    }
  }


  /*???*/

}   /* end of InerToDispKD() */




/* ------------------------------------------------------------------- */
static StatusET            /* ret : OK, W_EMPTYCLASS or E_MEMORY */ /*V1.06-b*/
EstimParaLaplace
(
  const double   *C_NK,      /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           Nk,         /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/
  const ModelSpecT* SpecP,  /* I : model specification */ /*V1.06-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 */ /*V1.05-b*/
  ModelParaT    *ParaP      /* O : estimated parameters */
)
/* ------------------------------------------------------------------- */
{
  int      N = DataP->NbPts ;
  int      D = DataP->NbVars ;

  StatusET sts ;


  /* Common computation for any diagonal gaussian model with missing data */
  sts = CommonLaplaceDiag
    ( DataP->PointsM, DataP->SortPos_ND, N, D, C_NK, Nk, MissMode, 
      ParaP->Disp_KD, 
      ParaP->Center_KD, 
      EmptyK_P, ParaP->NbObs_K, ParaP->NbObs_KD, ParaP->Iner_KD ) ;

  /* Dispersion is computed depending on dispersion model */
  InerToDisp( SpecP->ClassDisper, N, Nk, D,SpecP, ParaP->NbObs_K, ParaP->NbObs_KD, 
	      ParaP->NbObs_KDD,ParaP->Iner_KD, MissMode, 
	      ParaP->Disp_KD, & sts ) ;

  return sts ;

  /*???*/
}   /* end of EstimParaLaplace() */


/*V1.06-f*/
/* ------------------------------------------------------------------- */
static StatusET          /* Return status : OK, EMPTY or MEMORY */
CommonLaplaceDiag
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const int*    Sort_ND,       /* I : sorted index / variable (N,D) */
  int           N,             /* I : nb of observation vectors */
  int           D,             /* I : nb of variables */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  int           K,             /* I : number of clusters */
  MissET        Miss,          /* I : how to treat missing data */
  const double*  OldDisp_KD,    /* I : old dispersions (K,D) */

  double*        OldNewCen_KD,  /* I/O : old then updated means (K,D) */

  int*          EmptyK_P,      /* O : index of empty class (0 or 1 ..K) */
  double*        N_K,           /* O : size of a class (K) */
  double*        N_KD,          /* O : size of a class and variable (K,D) */
  double*        Iner_KD        /* O : inertia of a class/variable (K,D,D) */
 )
/* ------------------------------------------------------------------- */
{
  const char* func = "CommonLaplaceDiag" ;

  double*   oldcent_KD ;    /* saved previous centers (K,D), local alloc */

  StatusET sts ;           /* return status */


  /* Allocate local structures */
  if ((oldcent_KD   = GenAlloc( K * D, sizeof(double), 1, func, "oldcent_KD" )) == NULL)
      fprintf(stderr,"impossible to allocate oldcent_KD in function CommonLaplaceDiag\n")  ;

  /* Save old mean value */
  memcpy( oldcent_KD , OldNewCen_KD , K * D * sizeof( double ) ) ;

  EstimSizes( X_ND, C_NK, N, D, K, 
	      N_K, N_KD ) ;

  sts = EstimLaplaceCenters( X_ND, Sort_ND, C_NK, N, D, K, Miss, 
			     N_K, N_KD, oldcent_KD, OldDisp_KD, 
			     OldNewCen_KD, EmptyK_P ) ;

  EstimLaplaceIner( X_ND, C_NK, N, D, K, Miss, 
		    N_K, N_KD, oldcent_KD, OldDisp_KD, OldNewCen_KD, 
		    Iner_KD ) ;

  /* Free local structures */
  GenFree( oldcent_KD ) ;

  return sts ;

  /*???*/
}   /* end of CommonLaplaceDiag() */


/* ------------------------------------------------------------------- */
static void EstimSizes
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  const int     N,             /* I : nb of observation vectors */
  const int     D,             /* I : nb of variables */
  const int     K,             /* I : number of clusters */
  double*        N_K,           /* O : size of a class (K) */
  double*        N_KD           /* O : size of a class and variable (K,D) */  
 )
/* ------------------------------------------------------------------- */
{
  int      h ;             /* current class    0..K-1 */
  int      j ;             /* current variable 0..D-1 */
  int      i ;             /* current object   0..N-1 */

  
  /* For each class h and variable j */
  for ( h = 0 ; h < K ; h ++ ) {

      for ( j = 0 ; j < D ; j ++ ) {

	  /* Initialize the sum[i] quantities to 0 */
	  N_K[ h ]             = 0.0 ;
	  N_KD[ _HJ ]          = 0.0 ;

	  /* For each object i, increment class size and
	     eventually observed data size if xij not nan */
	  for ( i = 0 ; i < N ; i ++ ) {

	      double cih = C_NK[ _IH ] ;
	      double xij = X_ND[ _IJ ] ;

	      N_K[ h ] += cih ;

	      if ( ! isnan( xij ) ) {
		  N_KD[ _HJ ]          += cih ;
	      }
	  } /* End  For each object i (all sums[i] are now done) */
      }
  }  /* End   for each class h and variable j */

}   /* end of EstimSizes() */



/* ------------------------------------------------------------------- */
static StatusET          /* Return status : OK, EMPTY */
EstimLaplaceCenters
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const int*    Sort_ND,       /* I : sorted index / variable (N,D) */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  const int     N,             /* I : nb of observation vectors */
  const int     D,             /* I : nb of variables */
  const int     K,             /* I : number of clusters */
  const MissET  Miss,          /* I : how to treat missing data */
  const double*  N_K,           /* I : size of a class (K) */
  const double*  N_KD,          /* I : size of a class and variable (K,D) */  
  const double*  OldCen_KD,     /* I : old centers */
  const double*  OldDisp_KD,    /* I : old dispersions (K,D) */

  double*        NewCen_KD,     /* O : updated centers (K,D) */
  int*          EmptyK_P       /* O : index of empty class (0 or 1 ..K) */
 )
/* ------------------------------------------------------------------- */
{
  StatusET sts ;           /* return status */

  int      h ;             /* current class    0..K-1 */
  int      j ;             /* current variable 0..D-1 */

  int      imed ;          /* index of median of observed values */
  double    medval ;        /* median value (eventually midway) */
  double    cumwei ;        /* cumulated weights until median observation */


  /* Set no empty class by default */
  (*EmptyK_P) = 0 ;
  sts         = STS_OK ;


  /* For each class h and variable j */
  for ( h = 0 ; h < K ; h ++ ) {

    for ( j = 0 ; j < D ; j ++ ) {

      /* If this class size > 0 */
      if ( N_K[ h ] > EPSILON ) {
	
	/* First, find median position of observed values 
	 */
	ComputeMedian( X_ND, Sort_ND, C_NK, N, D, K, h, j, N_KD[ _HJ ],
		       & imed, & cumwei, & medval ) ;

	/* 
	 * Compute center from observed median, depending on presence
	 * of missing data, REPLACE or IGNORE missing mode 
	 */

	if ( N_KD[ h * D + j ] == N_K[ h ] ) /* no missing data */ {
	  NewCen_KD[ h * D + j ] = medval ;
	}
	else if ( N_KD[ h * D + j ] >= EPSILON ) /* some missing data */ {
	   if ( Miss == MISSING_REPLACE ) {

	     NewCen_KD[ h * D + j ] = 
	       FindMinInerLaplaceEM( X_ND, Sort_ND, C_NK, N, D, K, h, j, 
				     imed, cumwei, OldCen_KD[ h * D + j ], 
				     OldDisp_KD[h*D*(D+1) +j*(2*D+1-j)/2 ],
				     N_KD[ h * D + j ], 
				     N_K[ h ] - N_KD[ h * D + j ] ) ;

	     /* 
		NewCen_KD[ h * D + j ] =  
		( N_KD[ h * D + j ] * medval + 
		( N_K[ h ] - N_KD[ h * D + j ] ) * OldCen_KD[ h * D + j ] )
		/ N_K[ h ] ;
		*/
	   }
	   else  /* assume Miss == MISSING_IGNORE */ {
	     NewCen_KD[ h * D + j ] = medval ;
	   }
	}
	else /* N_KD[ h * D + j ] < EPSILON : all data missing */ {
	  NewCen_KD[ h * D + j ] = OldCen_KD[ h * D + j ] ;
	}

      }
      else /* then this class size == 0 */ {
	NewCen_KD[ h * D + j ] = OldCen_KD[ h * D + j ] ;
	/* signal empty class and which one */
	sts = STS_W_EMPTYCLASS ;
	(*EmptyK_P) = h + 1 ;
      }

    }
  }
  /* end  For each class h and variable j */

  return sts ;
  /*???*/
}   /* end of EstimLaplaceCenters() */



/* ------------------------------------------------------------------- */
static void ComputeMedian
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const int*    Sort_ND,       /* I : sorted index / variable (N,D) */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  const int     N,             /* I : nb of observation vectors */
  const int     D,             /* I : nb of variables */
  const int     K,             /* I : number of clusters */
  const int     H,             /* I : current cluster */
  const int     J,             /* I : current variable */
  const double   totwei,        /* I : total weight */
  int*          ImedP,         /* O : index of median of observed values */
  double*        CumweiP,       /* O : cumulated weights until median obs. */
  double*        MedvalP        /* O : median value (eventually midway) */
 )
/* ------------------------------------------------------------------- */
{
  double    halfwei = totwei / 2 ;
  int      i ;             /* current index in ascending order  0..N-1 */
  int      pos ;           /* actual position of current object 0..N-1 */
  int      posmed ;        /* actual position of median 0..N-1 */
  int      posnext ;       /* actual position of next to median 0..N-1 */

  
  /* Scan the data in ascending order (skip NaN values), and find position 
     where the cumulated weights >= half total weight */
  for ( i = 0, 
	  (*CumweiP) = 0.0 ;
	( i < N ) &&
	  (*CumweiP) < halfwei ;
	i ++ ) {

    pos = Sort_ND[ i * D + J ] ;
    if ( !isnan( X_ND[ pos * D + J ] ) ) 
      (*CumweiP) += C_NK[ pos * K + H ] ;

  }  /* At this point, (*CumweiP) >= halfwei */

  (*ImedP) = i - 1 ;
  posmed = Sort_ND[ (*ImedP) * D + J ] ;

  /* Median value depends if = or > */
  if ( (*CumweiP) > halfwei + EPSILON ) /* > : median value is here*/ {

    (*MedvalP) = X_ND[ posmed * D + J ] ;
  }
  else  /* = : median value is midway to next non nan
 observation */ {

    for ( i = (*ImedP) + 1 ;
	  isnan( X_ND[ Sort_ND[ i * D + J ] * D + J ] ) ||
	    ( C_NK[ Sort_ND[ i * D + J ] * K + H ] < EPSILON ) ;
	  i ++ ) {
    }
    posnext = Sort_ND[ i * D + J ] ;
    (*MedvalP) = 0.5 * ( X_ND[ posmed * D + J ] + X_ND[ posnext * D + J ] ) ;
  }

}   /* end of ComputeMedian() */




/* ------------------------------------------------------------------- */
static double FindMinInerLaplaceEM  /* ret: minimizer of expected inertium */
 (
  const double*  X_ND,          /* I : data matrix (N,D) */
  const int*    Sort_ND,       /* I : sorted index / variable (N,D) */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  const int     N,             /* I : nb of observation vectors */
  const int     D,             /* I : nb of variables */
  const int     K,             /* I : number of clusters */
  const int     H,             /* I : current cluster */
  const int     J,             /* I : current variable */
  const int     Imed,          /* I : median position of observed data */
  const double   Cumwei,        /* I : cumulated weight until Xmed */
  const double   Cen0,          /* I : previous center */
  const double   Disp0,         /* I : previous dispersion */
  const double   Nhj,           /* I : number of observations */
  const double   Nmis           /* I : number of missing data */
 )
/* ------------------------------------------------------------------- */
{
  double   xmed ;       /* observation at median position */
  double   weidif ;     /* (before + med) - after or (med + after) - before */
  double   weibefore ;  /* cumulated weight before median position */
  int     direction ;  /* 1 if xmed < cen0, -1 otherwise */

  double   a ;       /* lower bound of current interval */
  double   b ;       /* upper bound of current interval */
  double   intwei ;  /* cumulated intermediate weight */
  double   d ;       /* current derivative R'(.) */
  double   res ;     /* value to return (minimizes expected inertia) */
  int     i ;       /* current index in ascending order  Ifirst..Ilast */

  int     sts ;     /* 0 : R' < 0 until now
		       1 : R'(a) >= 0
		       2 : R'(a) < 0 and R'(b) >= 0 */

  xmed = X_ND[ Sort_ND[ Imed * D + J ] * D + J ] ;

  if ( xmed < Cen0 - EPSILON ) {
    direction =  1 ;
    weidif    =  Cumwei - ( Nhj - Cumwei ) ;
  }
  else if ( xmed >  Cen0 + EPSILON ) {
    direction =  -1 ;
    weibefore =  Cumwei - C_NK[ Sort_ND[ Imed * K + H ] * K + H ] ;
    weidif    =  (Nhj - weibefore) - weibefore ;
  }
  else   /* xmed == Cen0, no search necessary */ {
    return Cen0 ;
  }


  /* Start at median position */
  a      = xmed ;
  intwei = 0.0 ;
  sts    = 0 ; 

  if ( (d = DerivInerDir( a, weidif, intwei, Nmis, Cen0, Disp0 )) >= 0 ) 
    /* R'(a) >= 0,  minimum at a */ {
    sts = 1 ;  
  }
  else  /* R'(a) < 0, look at b */ {

    for ( i = Imed + direction ;

	  ( sts == 0 ) && 
	    ( (X_ND[ Sort_ND[ i * D + J ] * D + J ] - Cen0) * direction < 0 ) 
	    && ( i >= 0 ) && ( i < N ) ;      /* "safety belt" */

	  i = i + direction ) 
      /* For each intermediate point */ {

      if ( ( ! isnan( X_ND[ Sort_ND[ i * D + J ] * D + J ] ) )
	   && ( C_NK[ Sort_ND[ i * D + J ] * K + H ] > EPSILON ) ) {

	b = X_ND[ Sort_ND[ i * D + J ] * D + J ] ;
	if ( (d = DerivInerDir( b, weidif, intwei, Nmis, Cen0, Disp0 )) >= 0 ) 
	  /* R'(b) >= 0, and R'(a) < 0 */ {

	  sts = 2 ;
	}
	else /* R'(b) < 0 */ {

	  /* Start checking new interval */
	  a = b ;
	  intwei += 2 * C_NK[ Sort_ND[ i * D + J ] * K + H ] ;
	  if ( (d = DerivInerDir( a, weidif, intwei, Nmis, Cen0, Disp0 )) >= 0)
	    /* R'(a) >= 0,  minimum at a */ {

	    sts = 1 ;
	  }
	}
      }
    }    /* end for each intermediate point */
  }

  if ( sts == 0 ) 
    /* R'(b) < 0, R'(a) < 0 => check for b = Cen0 */ {
    b = Cen0 ;
    if ( (d = DerivInerDir( b, weidif, intwei, Nmis, Cen0, Disp0 )) >= 0 ) 
      /* R'(b) >= 0, and R'(a) < 0 */ {

      sts = 2 ;
    }
  }

  switch( sts ) {
  case 0:
    /* R'(a) < 0, R'(b) < 0 => Cen0 is the minimizer */
    res = Cen0 ;
    break ;

  case 1:
    /* R'(a) >= 0 => a is the minimizer */
    res = a ;
    break ;

  default:
    /* R'(a) < 0, R'(b) >= 0 => 
       y in [a,b] minimizing R is solution of :
       | y - Cen0 | = - Disp0 log( 1 - ( weidif - intwei ) / Nmis )
       */
    res = Cen0 + direction * Disp0 * log( 1 - ( weidif + intwei ) / Nmis ) ;
  }

  return res ;
  /*???*/
}   /* end of FindMinInerLaplaceEM() */


/* ------------------------------------------------------------------- */
static double DerivInerDir      /* ret: +/- R'(Y), R expected inertia */ 
 (
  const double   Y,             /* I : point at which to compute R' */
  const double   Weidif,        /* I : "after+med-bef" or "bef+med-after" */
  const double   Intwei,        /* I : cumulated intermediate weight */
  const double   Nmis,          /* I : number of missing data */
  const double   Cen0,          /* I : previous center */
  const double   Disp0          /* I : previous dispersion */
 )
/*\
     This function returns 
       R'(Y)   if Xmed < Cen0
     - R'(Y)   if Xmed > Cen0
     Testing the positiveness of this function thus tests R'(Y) > 0
     in the first case and R'(Y) < 0 in the second case.
\*/
/* ------------------------------------------------------------------- */
{

  double d = Weidif + Intwei 
    - Nmis * ( 1 - exp( - fabs( Y - Cen0 ) / Disp0 ) ) ;

  return d ;

  /*???*/
}   /* end of DerivIner() */




/* ------------------------------------------------------------------- */
static void EstimLaplaceIner
 ( 
  const double*  X_ND,          /* I : data matrix (N,D) */
  const double*  C_NK,          /* I : classification matrix (N,K) */
  const int     N,             /* I : nb of observation vectors */
  const int     D,             /* I : nb of variables */
  const int     K,             /* I : number of clusters */
  const MissET  Miss,          /* I : how to treat missing data */
  const double*  N_K,           /* I : size of a class (K) */
  const double*  N_KD,          /* I : size of a class and variable (K,D) */  
  const double*  OldCen_KD,     /* I : old centers */
  const double*  OldDisp_KD,    /* I : old dispersions (K,D) */
  const double*  NewCen_KD,     /* I : new centers */
  double*        Iner_KD        /* O : inertia of a class/variable (K,D,D) */  
 )
/* ------------------------------------------------------------------- */
{
  int      h ;             /* current class    0..K-1 */
  int      j ;             /* current variable 0..D-1 */
  int      i ;             /* current object   0..N-1 */

  
  /* For each class h and variable j */
  for ( h = 0 ; h < K ; h ++ ) {

      for ( j = 0 ; j < D ; j ++ ) {

	  /* Initialize the inertium to 0 */
	  Iner_KD[h*D*(D+1)/2+j*(2*D+1-j)/2 ] = 0.0 ;

	  /* For each non nan object i, increment class inertium */
	  for ( i = 0 ; i < N ; i ++ ) {

	      double cih = C_NK[ _IH ] ;
	      double xij = X_ND[ _IJ ] ;

	      if ( ! isnan( xij ) ) {
		Iner_KD[h*D*(D+1)/2+j*(2*D+1-j)/2] += 
                      cih * fabs( xij - NewCen_KD[ _HJ ] ) ;
	      }

	  } /* End  For each non-nan object i (all sums[i] are now done) */

	  
	  if ( ( N_KD[ _HJ ] < N_K[ h ] ) &&  ( Miss == MISSING_REPLACE ) )
	       /* If some missing data and replace mode */ {

	    Iner_KD[h*D*(D+1)/2+j*(2*D+1-j)/2   ] += 
	      ( N_K[ h ] - N_KD[ _HJ ] ) *
	      ( fabs( OldCen_KD[ _HJ ] - NewCen_KD[ _HJ ] ) +
		OldDisp_KD[h*D*(D+1)/2+j*(2*D+1-j)/2 ] * 
		exp( - fabs( OldCen_KD[ _HJ ] - NewCen_KD[ _HJ ] ) / 
		     OldDisp_KD[h*D*(D+1)/2+j*(2*D+1-j)/2] ) ) ; 
	  }

      }
  }  /* End   for each class h and variable j */
  
  /*???*/
}   /* end of EstimLaplaceIner() */




/*+++*/

#if 0


/* ------------------------------------------------------------------- */
StatusET                    /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaP_V_I
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
  int      k ;
  StatusET sts ;


  sts = ParaPkV_I( Cih_nk, DataP, K, MissMode, EmptyK_P, NoiseParaP ) ;

  for ( k = 0 ; k < K ; k ++ )
    {
        NoiseParaP->Pk[ k ] = 1.0 / K ;
    }
  return sts ;
}

/* ------------------------------------------------------------------- */
StatusET ParaPkV_I          /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{

    if ( DataP->NbMiss == 0 )
      return ParaPkV_I_Complete( Cih_nk, DataP, K, MissMode, 
				 EmptyK_P, NoiseParaP ) ;
    else
      return ParaPkV_I_Missing( Cih_nk, DataP, K, MissMode, 
				EmptyK_P, NoiseParaP ) ;

}   /* end of ParaPkV_I() */


/* ------------------------------------------------------------------- */
static StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaPkV_I_Complete
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
    int         h ;         /* class counter : 0 .. K - 1 */
    double       sumh_trTh ; /* sum over h of class h's dispersion matrix's trace */
    double       V ;         /* common volume = sum[h] trTh / (N * Nd) */
    int         nd  = DataP->NbVars ;   /* Nd : nb of variables */
    int         npt = DataP->NbPts ;    /* N : nb of points */

    /* For each class */
    for ( h = 0 , (*EmptyK_P) = 0, sumh_trTh = 0.0 ; 
          h < K ; 
          h ++ )
    {
        int     ipt ;
        int     d ;
        double   sumi_uih ;
        double   sumi_uih_yi2 ;

        /* 'fuzzy' cardinal : sizeh = sum[i]( uih ) */
        /* proportion :       ph    = 1 / K */
        /* mean :             mh(d) = sum[i]( uih * yi(d) ) / sizeh */
        /* trace of its dispersion matrix : 
           trTh = sum[i]( uih * sum[d] yi(d)? ) - sizeh * sum[d] mh(d)? */
        /* volume :           Vh = V = sum[h] trTh / (N * Nd) */

        /* Initialize means to zero */
        for ( d = 0 ; d < nd ; d ++ )
        {
            NoiseParaP->Mk[ ( h * nd ) + d ] = 0 ;
        }

        /* Do cumulated sums over all points */
        for ( ipt = 0, sumi_uih = 0.0, sumi_uih_yi2 = 0.0 ; 
              ipt < npt ; 
              ipt ++ )
        {
            double   sumd_uih_yid2 ;
            double   uih                             = Cih_nk[ ( ipt * K ) + h ] ;

            for ( d = 0, sumd_uih_yid2 = 0.0 ; 
                  d < nd ; 
                  d ++ )
            {
                double   yid   = DataP->PointsM[ (ipt * nd) + d ] ;

                NoiseParaP->Mk[ ( h * nd ) + d ]    += uih * yid ;
                sumd_uih_yid2                       += uih * yid * yid ;
            }

            sumi_uih        += uih ;
            sumi_uih_yi2    += sumd_uih_yid2 ;
        }

        /* Normalize the sums by the cardinal of the class */
        if ( sumi_uih != 0.0 )
        {
            double   sumd_mhd2 ;
            double   trTh ;
            double   invsizeh = 1 / sumi_uih ;

            for ( d = 0, sumd_mhd2 = 0.0 ; 
                  d < nd ; 
                  d ++ )
            {
                double   mhd         = NoiseParaP->Mk[ (h * nd) + d ] ;

                mhd                 *= invsizeh ;
                sumd_mhd2           += mhd * mhd ;

                NoiseParaP->Mk[ (h * nd) + d ]  = mhd ;
            }

            trTh                    = sumi_uih_yi2 - sumi_uih * sumd_mhd2 ;
            sumh_trTh               += trTh ;
        }
        else
        {
            (*EmptyK_P) = h + 1 ;
        }

        NoiseParaP->Pk[ h ] = sumi_uih / npt ;

        /* Set normalized variance matrix to identity */
        SetIdMatrix( nd , & NoiseParaP->Ck[ h*nd*nd ] ) ;  /*V1.03-a*/

    } /* end for ( h = 0 , (*EmptyK_P) = 0 ; h < K ; h ++ ) */

    V = sumh_trTh / ( npt * nd ) ;

    for ( h = 0 ; h < K ; h ++ )
    {
        NoiseParaP->Vk[ h ] = V ;
    }

    if ( (*EmptyK_P) == 0 )
      return STS_OK ;
    else
      return STS_W_EMPTYCLASS ;

}   /* end of ParaPkV_I_Complete() */



/* ------------------------------------------------------------------- */
static StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaPkV_I_Missing
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
  const char* func = "ParaPkV_I_Missing" ;

  int      N = DataP->NbPts ;
  int      D = DataP->NbVars ;

  int      h ;         /* current class    0..K-1 */
  int      j ;         /* current variable 0..D-1 */

  double*   nh_k ;      /* size / class (K) local alloc (unused) */
  double*   nhj_kd ;    /* size / obs class & var (K,D) local alloc (unused)*/
  double*   disphj_kd ; /* dispersion / class & var (K,D) local alloc */
  double    vol ;       /* common volume */

  StatusET sts ;


  /* Allocate local data */
  nh_k      = GenAlloc( K,     sizeof( double ), 1, func, "nh_k" ) ;
  nhj_kd    = GenAlloc( K * D, sizeof( double ), 1, func, "nhj_kd" ) ;
  disphj_kd = GenAlloc( K * D, sizeof( double ), 1, func, "disphj_kd" ) ;

 if ( (nh_k == NULL) || (nhj_kd == NULL) || (disphj_kd == NULL) )
    fprintf(stderr," impossible to allocate variables in function ParaPkV_I_Missing\n");

  /* Common computation for any diagonal gaussian model with missing data */
  sts = CommonGaussDiagMissing
    ( DataP->PointsM, DataP->NbPts, DataP->NbVars, Cih_nk, K, MissMode,
      NoiseParaP->Vk, NoiseParaP->Ck, 
      NoiseParaP->Mk, NoiseParaP->Pk, EmptyK_P, nh_k, nhj_kd, disphj_kd ) ;

  /* Compute common volume : 
     sum dispersions over the classes and variables 
     then divide by all/observed data size */
  
  vol = 0.0 ;

  for ( h = 0; h < K ; h ++ )
    {
      /* If not empty class */
      if ( nh_k[ h ] > 0 )
	{
	  for ( j = 0 ; j < D ; j ++ )
	    {
	      vol += disphj_kd[ _HJ ] ;
	    }
	}
    }

  if ( MissMode == MISSING_REPLACE )
    vol /= ( N * D ) ;
  else
    vol /= ( N * D - DataP->NbMiss ) ;

  /* For each class h */
  for ( h = 0; h < K ; h ++ )
    {
      /* Set its volume to the common volume */
      NoiseParaP->Vk[ h ] = vol ;

      /* Set its normalized covariance to identity */
      SetIdMatrix( D, & NoiseParaP->Ck[ h * D * D ] ) ;
    }
      
  /* Free local data */
  GenFree( disphj_kd ) ;
  GenFree( nhj_kd ) ;
  GenFree( nh_k ) ;

  return sts ;
}   /* end of ParaPkV_I_Missing() */



/* ------------------------------------------------------------------- */
StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaP_VkI
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
  StatusET sts ;
  int k ;

  sts = ParaPkVkI( Cih_nk, DataP, K, MissMode, EmptyK_P, NoiseParaP ) ;

  for ( k = 0 ; k < K ; k ++ )
    {
        NoiseParaP->Pk[ k ] = 1.0 / K ;
    }
  return sts ;
}


/* ------------------------------------------------------------------- */
StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaPkVkI
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{

    if ( DataP->NbMiss == 0 )
      return ParaPkVkI_Complete( Cih_nk, DataP, K, MissMode, 
				 EmptyK_P, NoiseParaP ) ;
    else
      return ParaPkVkI_Missing( Cih_nk, DataP, K, MissMode, 
				EmptyK_P, NoiseParaP ) ;

}   /* end of ParaPkV_I() */



/* ------------------------------------------------------------------- */
static StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaPkVkI_Complete
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
    int         k ;

    /* For each class */
    for ( k = 0 , (*EmptyK_P) = 0 ; k < K ; k ++ )
    {
        int     ipt ;                  
        int     id ;
        int     nd  = DataP->NbVars ;
        int     npt = DataP->NbPts ;    /* = n nb points */
        double   sizek = 0.0 ;            /* = nk fuzzy cardinal [0,n] */
        double   trWk = 0.0 ;

        /* Compute its 'fuzzy' cardinal : nk = sum[i=1,n]( uik ) */
        /* Compute its proportion : pk = nk / n */
        /* Compute its mean (mk) : mk = sum[i=1,n]( uik * Y(i,:) ) / nk */
        /* Compute its intra-class inertia matrix : 
           Wk = sum[i=1,n]( uik * ( Y(i,:) - mk ) * ( Y(i,:) - mk )' ) / nk */
        /* Compute its volume : Vk = tr(Wk) / d */

        /* Initialize means to zero */
        for ( id = 0 ; id < nd ; id ++ )
        {
            NoiseParaP->Mk[ (k * nd) + id ] = 0 ;
        }

        /* Do cumulated sums over all points */
        for ( ipt = 0 ; ipt < npt ; ipt ++ )
        {
            double   uik = Cih_nk[ (ipt * K) + k ] ;

            sizek += uik ;
            for ( id = 0 ; id < nd ; id ++ )
            {
                double   yid = DataP->PointsM[ (ipt * nd) + id ] ;
                double   mkd = NoiseParaP->Mk[ (k * nd) + id ] ;

                mkd         = mkd + uik * yid ;
                trWk        = trWk + uik * yid * yid ;

                NoiseParaP->Mk[ (k * nd) + id ] = mkd ;
            }
        }

        /* Normalize the sums by the cardinal of the class */
        if ( sizek != 0.0 )
        {
            trWk /= sizek ;

            for ( id = 0 ; id < nd ; id ++ )
            {
                double   mkd = NoiseParaP->Mk[ (k * nd) + id ] ;

                mkd         = mkd / sizek ;
                trWk        = trWk - ( mkd * mkd ) ;

                NoiseParaP->Mk[ (k * nd) + id ] = mkd ;
            }
            NoiseParaP->Vk[ k ] = trWk / nd ;
        }
        else
        {
            (*EmptyK_P) = k + 1 ;
        }

        NoiseParaP->Pk[ k ] = sizek / npt ;

        /* Set normalized variance matrix to identity */
        SetIdMatrix( nd , & NoiseParaP->Ck[ k*nd*nd ] ) ;  /*V1.03-a*/

    } /* end for ( k = 0 , (*EmptyK_P) = 0 ; k < K ; k ++ ) */

    if ( (*EmptyK_P) == 0 )
      return STS_OK ;
    else
      return STS_W_EMPTYCLASS ;

}   /* end of ParaPkVkI_Complete() */


/* ------------------------------------------------------------------- */
static StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaPkVkI_Missing
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
  const char* func = "ParaPkVkI_Missing" ;

  int      N = DataP->NbPts ;
  int      D = DataP->NbVars ;

  int      h ;         /* current class    0..K-1 */
  int      j ;         /* current variable 0..D-1 */

  double*   nh_k ;      /* size / class (K) local alloc */
  double*   nhj_kd ;    /* size / obs class and variable (K,D) local alloc */
  double*   disphj_kd ; /* dispersion / class and variable (K,D) local alloc */

  double    sumj_nhj ;  /* size of obs class */

  StatusET sts ;


  /* Allocate local data */
  nh_k      = GenAlloc( K,     sizeof( double ), 1, func, "nh_k" ) ;
  nhj_kd    = GenAlloc( K * D, sizeof( double ), 1, func, "nhj_kd" ) ;
  disphj_kd = GenAlloc( K * D, sizeof( double ), 1, func, "disphj_kd" ) ;

 if ( (nh_k == NULL) || (nhj_kd == NULL) || (disphj_kd == NULL) )
    fprintf(stderr," impossible to allocate variables in function ParaPkVkI_Missing\n");

  /* Common computation for any diagonal gaussian model with missing data */
  sts = CommonGaussDiagMissing
    ( DataP->PointsM, DataP->NbPts, DataP->NbVars, Cih_nk, K, MissMode,
      NoiseParaP->Vk, NoiseParaP->Ck, 
      NoiseParaP->Mk, NoiseParaP->Pk, EmptyK_P, nh_k, nhj_kd, disphj_kd ) ;

  /* For each class h */
  for ( h = 0; h < K ; h ++ )
    {
      /* If not empty class */
      if ( nh_k[ h ] > 0 )
	{
	  /* Reestimate its volume: sum dispersions over the variables 
	     then divide by all/observed data size */
	  sumj_nhj = 0 ;
	  NoiseParaP->Vk[ h ] = 0.0 ;  /* use as sum first */

	  for ( j = 0 ; j < D ; j ++ )
	    {
	      sumj_nhj            += nhj_kd[ _HJ ] ;
	      NoiseParaP->Vk[ h ] += disphj_kd[ _HJ ] ;
	    }

	  if ( MissMode == MISSING_REPLACE )
	    NoiseParaP->Vk[ h ] /= nh_k[ h ] ;
	  else
	    NoiseParaP->Vk[ h ] /= sumj_nhj ;
	}
      else /* class h is empty */
	{
	  NoiseParaP->Vk[ h ] = 0 ;
	}

      /* Set its normalized covariance to identity */
      SetIdMatrix( D, & NoiseParaP->Ck[ h * D * D ] ) ;
    }

  /* Free local data */
  GenFree( disphj_kd ) ;
  GenFree( nhj_kd ) ;
  GenFree( nh_k ) ;

  return sts ;
}



/*V1.05-d*/
/* ------------------------------------------------------------------- */
static StatusET                /* Returns : OK, W_EMPTYCLASS or E_MEMORY */
CommonGaussDiagMissing
 (
  const double*  Xij_nd,        /* I : data matrix (N,D) */
  int           N,             /* I : nb of observation vectors */
  int           D,             /* I : nb of variables */
  const double*  Cih_nk,        /* I : classification matrix (N,K) */
  int           K,             /* I : number of clusters */
  MissET        Miss,          /* I : how to treat missing data */
  const double*  OldVh_k,       /* I : old volumes (K) */
  const double*  OldChjj_kdd,   /* I : old normalized covariances (K,D,D) */

  double*        OldNewMhj_kd,  /* I/O : old then updated means (K,D) */
  double*        NewPh_k,       /* O : updated proportions (K) */
  int*          EmptyK_P,      /* O : index of empty class (0 or 1 ..K) */
  double*        Nh_k,          /* O : size of a class (K) */
  double*        Nhj_kd,        /* O : size of a class and variable (K,D) */
  double*        Disphj_kd      /* O : dispersion of a class/variable (K,D) */
 )
/*\

   This function does the common computation that is needed at the
   M-step, in all the diagonal Gaussian models with missing data.

   The routine takes as input a data matrix Xij_nd (possibly
   containing NaN values), a classification matrix Cih_nk, and the
   kind of missing data processing Miss (REPLACE or IGNORE) ; and also
   the old volumes OldVh_k, covariance matrices OldChjj_kdd, and means
   OldNewMhj_kd of the classes, to be used in the REPLACE mode.

   It produces as output the reestimated mean vectors in OldNewMhj_kd;
   reestimated proportions in NewP_k; the last encountered empty class
   in EmptyK_P (or 0 if no empty class); the {sum[i] cih} in Nh_k, the
   {sum[i] cih rij} in Nhj_kd (number of present observations in class h
   and variable j); and the "missing data dispersion" of each
   class/variable in Disphj_kd (i.e. the missing data equivalent of
   {sum[i] Cih (xij - mhj)^2}).

\*/
/* ------------------------------------------------------------------- */
{
  const char* func = "CommonGaussDiagMissing" ;

  double*   sumdata_hj_kd ;    /* sum[i] cih rij xij   (K,D), local alloc */
  double*   sumsquare_hj_kd ;  /* sum[i] cih rij xij^2 (K,D), local alloc */
  double*   oldmean_hj_kd ;    /* saved previous means (K,D), local alloc */
  int      h ;                /* current class    0..K-1 */
  int      j ;                /* current variable 0..D-1 */
  int      i ;                /* current object   0..N-1 */
  StatusET sts ;              /* return status */


  sts = STS_OK ;

  /* Allocate local structures */
  sumdata_hj_kd   = GenAlloc( K * D, sizeof(double), 1, func, "sumdata" ) ;
  sumsquare_hj_kd = GenAlloc( K * D, sizeof(double), 1, func, "sumsquare" );
  oldmean_hj_kd   = GenAlloc( K * D, sizeof(double), 1, func, "oldmean" ) ;

 if ( (sumdata_hj_kd == NULL) || ( sumsquare_hj_kd== NULL) || ( oldmean_hj_kd== NULL) )
    fprintf(stderr," impossible to allocate variables in function CommonGaussDiagMissing\n");

  /* Save old mean value */
  memcpy( oldmean_hj_kd , OldNewMhj_kd , K * D * sizeof( double ) ) ;

  /* Set no empty class by default */
  (*EmptyK_P) = 0 ;

  /* For each class h */
  for ( h = 0 ; h < K ; h ++ )
    {
      /* For each variable j */
      for ( j = 0 ; j < D ; j ++ )
	{
	  /* Initialize the sum[i] quantities to 0 */
	  Nh_k[ h ]               = 0.0 ;
	  Nhj_kd[ _HJ ]           = 0.0 ;
	  sumdata_hj_kd[ _HJ ]    = 0.0 ;
	  sumsquare_hj_kd[ _HJ ]  = 0.0 ;
	  OldNewMhj_kd[ _HJ ]     = 0.0 ; /* use Mhj as sum first */

	  /* For each object i */
	  for ( i = 0 ; i < N ; i ++ )
	    {
	      double cih = Cih_nk[ _IH ] ;
	      double xij = Xij_nd[ _IJ ] ;

	      /* Increment class size */
	      Nh_k[ h ] += cih ;

	      /* If this object's j variable is observed */
	      if ( ! isnan( xij ) )
		{
		  /* Increment all sum[i] of observed quantities */
		  Nhj_kd[ _HJ ]           += cih ;
		  sumdata_hj_kd[ _HJ ]    += cih * xij ;
		  sumsquare_hj_kd[ _HJ ]  += cih * xij * xij ;
		}
	      /* Else xij is missing, don't change sum[i] of observed data */
	    }
	  /* End  For each object i (all sums[i] are now done) */

	  /* If this class size > 0 */
	  if ( Nh_k[ h ] > 0 )
	    {
	      /* 
	       * Compute means Mhj and dispersions Dhj from computed sums,
	       * depending on REPLACE or IGNORE missing mode 
	       */
	      if ( Miss == MISSING_REPLACE )
		{
		  /* 
		   * new_mhj = 
		   *  ( sum[i] cih rij xij + ( nh - nhj ) old_mhj ) / nh 
		   *
		   * disp_hj = 
		   *  "observed dispersion"_hj +
		   *  ( nh - nhj ) [ (new_mhj - old_mhj)^2 + old_vh old_chjj ]
		   */
		  double obsdisphj ;

		  OldNewMhj_kd[ _HJ ] = 
		    ( sumdata_hj_kd[ _HJ ] + 
		      ( Nh_k[ h ] - Nhj_kd[ _HJ ] ) * oldmean_hj_kd[ _HJ ] )
		    / Nh_k[ h ] ;

		  /*
		   * "observed dispersion"_hj = 
		   *      sum[i] cih rij xij^2 - nhj new_mhj^2 
		   */
		  obsdisphj = sumsquare_hj_kd[ _HJ ] - 
		    Nhj_kd[ _HJ ] * sqr( OldNewMhj_kd[ _HJ ] ) ;

		  Disphj_kd[ _HJ ] = 
		    obsdisphj +
		    ( Nh_k[ h ] - Nhj_kd[ _HJ ] ) *
		    ( sqr( OldNewMhj_kd[ _HJ ] - oldmean_hj_kd[ _HJ ] ) +
		      OldVh_k[ h ] * OldChjj_kdd[h*D(D+1)/2+j*(2*D+1-j)/2 ] ) ;
		}
	      else  /* assume Miss == MISSING_IGNORE */
		{
		  double obsdisphj ;

		  /* 
		   * new_mhj = 
		   *  . if  nhj <> 0,  ( sum[i] cih rij xij ) / nhj  
		   *  . if  nhj == 0,  old_mhj
		   *
		   * disp_hj = 
		   *  "observed dispersion"_hj 
		   */

		  if ( Nhj_kd[ _HJ ] > 0 )
		    OldNewMhj_kd[ _HJ ] = sumdata_hj_kd[ _HJ ] / Nhj_kd[ _HJ ];
		  else
		    OldNewMhj_kd[ _HJ ] = oldmean_hj_kd[ _HJ ] ;

		  /*
		   * "observed dispersion"_hj = 
		   *      sum[i] cih rij xij^2 - nhj new_mhj^2 
		   */
		  obsdisphj = sumsquare_hj_kd[ _HJ ] - 
		    Nhj_kd[ _HJ ] * sqr( OldNewMhj_kd[ _HJ ] ) ;

		  Disphj_kd[ _HJ ] = obsdisphj ;
		}

	    } 
	  else /* then this class size == 0 */
	    {
	      /* signal empty class and which one */
	      sts = STS_W_EMPTYCLASS ;
	      (*EmptyK_P) = h + 1 ;
	    }
	}
      /* end  For each variable j */

      /* Update proportions as  ph = nh / n */
      NewPh_k[ h ] = Nh_k[ h ] / N ;
    }
  /* end  For each class h */


  /* Free local structures */
  GenFree( oldmean_hj_kd ) ;
  GenFree( sumsquare_hj_kd ) ;
  GenFree( sumdata_hj_kd  ) ;

  return sts ;

}  /*  end of CommonGaussDiagMissing() */





/* ------------------------------------------------------------------- */
void SetIdMatrix ( int Nd , double* M )
/* Set matrix M to identity */ /*V1.03-a*/
/* ------------------------------------------------------------------- */
{
  int l , c ;  /* line and column counters : 0..Nd-1 */

  for ( l = 0 ; l < Nd ; l ++ )
    {
      for ( c = 0 ; c < Nd ; c ++ )
        {
          if ( c == l )
            M[ (l * Nd) + c ] = 1.0 ;
          else
            M[ (l * Nd) + c ] = 0.0 ;
        }
    }
}


#endif
/* ------------------------------------------------------------------- */

/* ~~~~~~~~~~~~~~~~~~~~~~~~ END OF FILE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
