/*
 ************************************************************************
 *	    		    C math library
 * function ZEROIN - obtain a function zero within the given range
 *
 * Input
 *	double zeroin(ax,bx,f,tol)
 *	double ax; 			Root will be seeked for within
 *	double bx;  			a range [ax,bx]
 *	double (*f)(double x);		Name of the function whose zero
 *					will be seeked for
 *	double tol;			Acceptable tolerance for the root
 *					value.
 *					May be specified as 0.0 to cause
 *					the program to find the root as
 *					accurate as possible
 *
 * Output
 *	Zeroin returns an estimate for the root with accuracy
 *	4*EPSILON*abs(x) + tol
 *
 * Algorithm
 *	G.Forsythe, M.Malcolm, C.Moler, Computer methods for mathematical
 *	computations. M., Mir, 1980, p.180 of the Russian edition
 *
 *	The function makes use of the bissection procedure combined with
 *	the linear or quadric inverse interpolation.
 *	At every step program operates on three abscissae - a, b, and c.
 *	b - the last and the best approximation to the root
 *	a - the last but one approximation
 *	c - the last but one or even earlier approximation than a that
 *		1) |f(b)| <= |f(c)|
 *		2) f(b) and f(c) have opposite signs, i.e. b and c confine
 *		   the root
 *	At every step Zeroin selects one of the two new approximations, the
 *	former being obtained by the bissection procedure and the latter
 *	resulting in the interpolation (if a,b, and c are all different
 *	the quadric interpolation is utilized, otherwise the linear one).
 *	If the latter (i.e. obtained by the interpolation) point is
 *	reasonable (i.e. lies within the current interval [b,c] not being
 *	too close to the boundaries) it is accepted. The bissection result
 *	is used in the other case. Therefore, the range of uncertainty is
 *	ensured to be reduced at least by the factor 1.6
 *
 ************************************************************************
 */

/* Modifications by FC the 07-12-2001 :

   - Add the both functions GetParameter and SetParameter
   - Changed the profile of the zeroin function, adding ParameterP,
     to know which quantity shall be minimized : beta or potential no i.
   - Changed the zeroin body : replaced all the calls to beta by
     SetParameter.
*/

#include "math.h"
#include "semms_typ.h"
#include <stdio.h>      /* printf, ... */

double GetParameter(double* Beta, double* Singl, ParameterT ParameterP) {
  if (ParameterP.Type == BETA)
    return *Beta;
  else
    return Singl[ParameterP.No];
}

void SetParameter(double* Beta, double* Singl, ParameterT ParameterP,
		  double  Value) {
  if (ParameterP.Type == BETA)
    *Beta = Value;
  else
    Singl[ParameterP.No] = Value;
}

StatusET  ZeroIn
/* An estimate to the root	*/
(
 const double  ax,		   /* Left border | of the range	*/
 const double  bx,  		   /* Right border| the root is seeked */
 ComputeCritGrad* f,		   /* Function under investigation */
 const double  tol,		   /* Acceptable tolerance */
 const int     Npt,                /* nb of points */
 const int     Nk,                 /* nb of classes */
 double*       Beta,
 const double* Ztilde,
 double*       Singl,
 const double* PostM,
 const SpatialT*  SpatialP,        /* neiborhood system */
 PtNeighsT*    NeighsP,
 ParameterT    ParameterP          /* To know what is the good derivative
				      to use and know which parameter must
				      be updated */
 )
/* ------------------------------------------------------------------- */
{
  double a,b,c;				/* Abscissae, descr. see above	*/
  double fa = 0;			/* f(a)                 	*/
  double fb = 0;			/* f(b)				*/
  double fc = 0;			/* f(c)				*/
  int    nb = 0;                        /* Number of iterations         */

  a = ax;  b = bx;
  SetParameter(Beta,Singl,ParameterP,a);
  fa = f(Npt,Nk,Beta,Ztilde,Singl,PostM,SpatialP,NeighsP,ParameterP);
  SetParameter(Beta,Singl,ParameterP,b);
  fb = f(Npt,Nk,Beta,Ztilde,Singl,PostM,SpatialP,NeighsP,ParameterP);
  /*fprintf(stderr,"Valeurs initiales pour l'appel de zero_in : \n");
  fprintf(stderr,"fa : %f \n",fa);
  fprintf(stderr,"fb : %f \n",fb);*/
  c = a;   fc = fa;

  /*fprintf(stderr," \n f(0) %6.3f f(10) %6.3f ", fa, fb ); */

  //fprintf(stderr,"in zeroin ");
  if ((fb > 0 && fa > 0) || (fb < 0 && fa < 0))
    {
      //fprintf(stderr,"f(ax) and f(bx) do not have different signs,");
      //fprintf(stderr,"zeroin is aborting\n");
      if (fb > 0) SetParameter(Beta,Singl,ParameterP,b);
      if (fb < 0) SetParameter(Beta,Singl,ParameterP,a);
      //fprintf(stderr,"out zeroin\n");
      return(0);
    }

  for(;;)		/* Main iteration loop	*/
  {
    double prev_step = b-a;		/* Distance from the last but one*/
					/* to the last approximation	*/
    double tol_act;			/* Actual tolerance		*/
    double p;      			/* Interpolation step is calcu- */
    double q;      			/* lated in the form p/q; divi- */
                                        /* sion operations is delayed   */
                                        /* until the last moment	*/
    double new_step;      		/* Step at this iteration       */

    if (fabs(fc) < fabs(fb))
      {                                 /* Swap data for b to be the 	*/
	a = b;  b = c;  c = a;          /* best approximation		*/
	fa=fb;  fb=fc;  fc=fa;
      }
    tol_act  = 2*EPSILON*fabs(b) + tol/2;
    new_step = (c-b)/2;

    if((fabs(new_step) <= tol_act && fb <= 5) || fb == (double)0)
      {
	// fprintf(stderr,"Number of iterations : %d\n",nb);
	SetParameter(Beta,Singl,ParameterP,b);
        /* Acceptable approx. is found */
	// fprintf(stderr,"out zeroin\n");
	return 0;
      }
    if (nb > 50)
      {
	fprintf(stderr,"Number of iterations exceeded 50. Leaving zeroin.\n");
	SetParameter(Beta,Singl,ParameterP,b);
	return 0;
      }
    nb++;

    /* Decide if the interpolation can be tried	*/
    if (fabs(prev_step) >= tol_act	/* If prev_step was large enough*/
	&& fabs(fa) > fabs(fb)) 	/* and was in true direction,	*/
      {					/* Interpolatiom may be tried	*/
	register double t1,cb,t2;
	cb = c-b;
	if( a==c )			/* If we have only two distinct	*/
	  {
	    /* fprintf(stderr,"points linear interpolation\n"); */
	    t1 = fb/fa;			/* can only be applied		*/
	    p = cb*t1;
	    q = 1.0 - t1;
	  }
	else
	  {
	    /* fprintf(stderr,"Quadric inverse interpolation\n"); */
	    q = fa/fc;  t1 = fb/fc;  t2 = fb/fa;
	    p = t2 * ( cb*q*(q-t1) - (b-a)*(t1-1.0) );
	    q = (q-1.0) * (t1-1.0) * (t2-1.0);
	  }
	if (p>(double)0)		/* p was calculated with the op-*/
	  q = -q;			/* posite sign; make p positive	*/
	else				/* and assign possible minus to	*/
	  p = -p;			/* q				*/
	if (p < (0.75*cb*q-fabs(tol_act*q)/2) /* If b+p/q falls in [b,c]*/
	    && p < fabs(prev_step*q/2) )	/* and isn't too large	*/
	  new_step = p/q;			/* it is accepted	*/
					/* If p/q is too large then the	*/
					/* bissection procedure can 	*/
					/* reduce [b,c] range to more	*/
					/* extent			*/
      }

    if (fabs(new_step) < tol_act)
      {	                                /* Adjust the step to be not less*/
      if (new_step > (double)0)	        /* than tolerance		*/
	new_step = tol_act;
      else
	new_step = -tol_act;
      }

    a = b;  fa = fb;			/* Save the previous approx.	*/
    b += new_step;
    /* fprintf(stderr,"zerocrit=%f\n",b);  */
    SetParameter(Beta,Singl,ParameterP,b);
    //fprintf(stderr,"Calling f\n");
    fb = f(Npt,Nk,Beta,Ztilde,Singl,PostM,SpatialP,NeighsP,ParameterP);
                                        /* Do step to a new approxim.	*/
    // fprintf(stderr,"No calling f\n");
    if ((fb > 0 && fc > 0) || (fb < 0 && fc < 0))
      {                 		/* Adjust c for it to have a sign*/
	c = a;  fc = fa;                /* opposite to that of b	*/
      }
  } /* End For */
} /* End ZeroIn */

/* ------------------------------------------------------------------- */
/*  StatusET  ZeroIn		 */
/*   An estimate to the root	*/
/*  ( */
/*  const double ax,			 Left border | of the range	*/
/*  const double bx,  		 Right border| the root is seeked*/
/*  ComputeCritGrad* f,			 Function under investigation	*/
/*  const double tol,		 Acceptable tolerance		*/
/*  const int    Npt,                nb of points */
/*  const int    Nk,                 nb of classes */
/*  double*       Beta,   */
/*  const double* Ztilde, */
/*  const double* Singl, */
/*  const double* PostM,   */
/*  const SpatialT*  SpatialP,       neiborhood system */
/*  PtNeighsT*    NeighsP */
/*  ) */
/*   ------------------------------------------------------------------- */
/*  { */
/*    double a,b,c;				 Abscissae, descr. see above	*/
/*    double fa = 0;				 f(a)				*/
/*    double fb = 0;				 f(b)				*/
/*    double fc = 0;				 f(c)				*/

/*    a = ax;  b = bx;   */
/*    fa = f(Npt,Nk,a,Ztilde,Singl,PostM,SpatialP,NeighsP); */
/*    fb = f(Npt,Nk,b,Ztilde,Singl,PostM,SpatialP,NeighsP); */
/*    c = a;   fc = fa; */

/*    fprintf(stderr," \n f(0) %6.3f f(10) %6.3f ", fa, fb ); */

/*     if ((fb > 0 && fa > 0) || (fb < 0 && fa < 0) ) */
/*    { */
/*     fprintf(stderr,"f(ax) and f(bx) do not have different signs, zeroin is aborting\n"); */
/*     if (fb > 0) *Beta = b; */
/*     if (fb < 0)  *Beta = a; */
/*     return(0); */
/*     }  */

/*    for(;;)		 Main iteration loop	*/
/*    { */
/*      double prev_step = b-a;		 Distance from the last but one*/
/*  					 to the last approximation	*/
/*      double tol_act;			 Actual tolerance		*/
/*      double p;      			 Interpolation step is calcu- */
/*      double q;      			 lated in the form p/q; divi- */
/*    					 sion operations is delayed   */
/*   					 until the last moment	*/
/*      double new_step;      		 Step at this iteration       */

/*      if( fabs(fc) < fabs(fb) ) */
/*      {                         		 Swap data for b to be the 	*/
/*  	a = b;  b = c;  c = a;           best approximation		*/
/*  	fa=fb;  fb=fc;  fc=fa; */
/*      } */
/*      tol_act = 2*EPSILON*fabs(b) + tol/2; */
/*      new_step = (c-b)/2; */

/*      if( (fabs(new_step) <= tol_act && fb <= 1) || fb == (double)0 ) */
/*        {*Beta= b;	 */
/*           Acceptable approx. is found */
/*         return(0); */
/*        } */
/*      			 Decide if the interpolation can be tried	*/
/*      if( fabs(prev_step) >= tol_act	 If prev_step was large enough*/
/*  	&& fabs(fa) > fabs(fb) )	 and was in true direction,	*/
/*      {					 Interpolatiom may be tried	*/
/*  	register double t1,cb,t2; */
/*  	cb = c-b; */
/*  	if( a==c )			 If we have only two distinct	*/
/*  	{	 */
/*  	   fprintf(stderr,"points linear interpolation\n"); */
/*  	  t1 = fb/fa;			can only be applied		*/
/*  	  p = cb*t1; */
/*  	  q = 1.0 - t1; */
/*   	} */
/*  	else	       */
/*  	{ */
/*             fprintf(stderr,"Quadric inverse interpolation\n"); */
/*  	  q = fa/fc;  t1 = fb/fc;  t2 = fb/fa; */
/*  	  p = t2 * ( cb*q*(q-t1) - (b-a)*(t1-1.0) ); */
/*  	  q = (q-1.0) * (t1-1.0) * (t2-1.0); */
/*  	} */
/*  	if( p>(double)0 )		 p was calculated with the op-*/
/*  	  q = -q;			 posite sign; make p positive	*/
/*  	else				 and assign possible minus to	*/
/*  	  p = -p;			 q				*/

/*  	if( p < (0.75*cb*q-fabs(tol_act*q)/2)	 If b+p/q falls in [b,c]*/
/*  	    && p < fabs(prev_step*q/2) )	 and isn't too large	*/
/*  	  new_step = p/q;			 it is accepted	*/
/*  					 If p/q is too large then the	*/
/*  					 bissection procedure can 	*/
/*  					 reduce [b,c] range to more	*/
/*  					 extent			*/
/*      } */

/*      if( fabs(new_step) < tol_act ){	 Adjust the step to be not less*/
/*        if( new_step > (double)0 )	 than tolerance		*/
/*  	new_step = tol_act; */
/*        else */
/*  	new_step = -tol_act;} */

/*      a = b;  fa = fb;			 Save the previous approx.	*/
/*      b += new_step; */
/*       fprintf(stderr,"zerocrit=%f\n",b);  */
/*      fb = f(Npt,Nk,b,Ztilde,Singl,PostM,SpatialP,NeighsP);	 Do step to a new approxim.	*/
/*      if( (fb > 0 && fc > 0) || (fb < 0 && fc < 0) ) */
/*      {                 			 Adjust c for it to have a sign*/
/*        c = a;  fc = fa;                   opposite to that of b	*/
/*      } */
/*    } */



/*  } */






